import numpy as np
from PIL import Image
from hypergan.viewer import GlobalViewer
from hypergan.samplers.base_sampler import BaseSampler
import time

class YSampler(BaseSampler):
    def __init__(self, gan, samples_per_row=8):
        BaseSampler.__init__(self, gan, samples_per_row)
        self.z = None
        self.y = None
        self.x = None
        self.mask = None
        self.step = 0
        self.steps = 8
        self.target = None
        self.y_t = gan.y.sample
        self.y = gan.session.run(self.y_t)
        self.g=tf.get_default_graph()
        self.frames = gan.session.run(gan.frames)
        self.frames_t = gan.frames
        self.zs2, self.cs2 = gan.session.run([gan.zs[-1], gan.cs[-1]])
        self.zs2 = [self.zs2]
        self.cs2 = [self.cs2]
        self.zs_t = [gan.video_generator_last_z]
        self.cs_t = [gan.video_generator_last_c]
        self.zs = gan.session.run([gan.video_generator_last_z])
        self.cs = gan.session.run([gan.video_generator_last_c])
        self.i=0

    def _sample(self):
        gan = self.gan
        z_t = gan.uniform_distribution.sample
        g=tf.get_default_graph()
        with g.as_default():
            tf.set_random_seed(1)
            feed_dict = dict(zip(self.frames_t, self.frames))
            sample, *gstack = gan.session.run(gan.gs_next, feed_dict=feed_dict)
            self.frames = self.frames[1:] + [sample]

            feed_dict = dict(zip(self.frames_t, self.frames))
            if self.i % 100 == 0:
                if self.i == 0:
                    self.frames3 = gan.session.run(gan.frames)
            feed_dict = dict(zip(self.frames_t, self.frames3))
            print('c_drift', gan.session.run(gan.c_drift, feed_dict=feed_dict))

            if self.i % 100 == 0:
                self.frames2 = gan.session.run(gan.frames)
            feed_dict = dict(zip(self.frames_t, self.frames2))
            sample_reset = gan.session.run(gan.gs_next[0], feed_dict=feed_dict)
            self.frames2 = self.frames2[1:] + [sample_reset]

            if self.i % 100 == 0:
                self.prev_frames = gan.session.run(gan.frames)
            feed_dict = dict(zip(gan.frames[1:], [self.prev_frames[0]]+self.prev_frames[:-1]))
            prev_sample = gan.session.run(gan.gy.sample, feed_dict=feed_dict)
            self.prev_frames = [prev_sample] + self.prev_frames[:-1]

            feed_dict = dict(zip(self.zs_t, self.zs))
            feed_dict.update(dict(zip(self.cs_t,self.cs)))
            gx, z_sample, c_sample = gan.session.run([gan.generator.sample, gan.video_generator_last_zn, gan.video_generator_last_cn], feed_dict=feed_dict)
            self.zs = self.zs[1:] + [z_sample]
            self.cs = self.cs[1:] + [c_sample]

            if self.i % 500 == 0:
                self.zs2, self.cs2 = gan.session.run([gan.zs[-1], gan.cs[-1]])
                self.zs2 = [self.zs2]
                self.cs2 = [self.cs2]
            feed_dict = dict(zip(self.zs_t, self.zs2))
            feed_dict.update(dict(zip(self.cs_t,self.cs2)))
            gc, z_sample, c_sample = gan.session.run([gan.generator.sample, gan.video_generator_last_zn, gan.video_generator_last_cn], feed_dict=feed_dict)
            self.zs2 = self.zs2[1:] + [z_sample]
            self.cs2 = self.cs2[1:] + [c_sample]

            feed_dict = dict(zip(self.zs_t, self.zs))
            x = gan.session.run(gan.generator.sample)
            self.i+=1
            return {
                'generator': np.hstack([sample,  sample_reset, gx,gc])
            }




from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from tensorflow.python.ops import control_flow_ops
from tensorflow.python.ops import math_ops
from tensorflow.python.ops import state_ops
from tensorflow.python.framework import ops
from tensorflow.python.training import optimizer
import hyperchamber as hc
import inspect
import numpy as np
from hypergan.train_hooks.base_train_hook import BaseTrainHook

class WeightConstraintTrainHook(BaseTrainHook):
  def after_create(self):
    self.max = self.gan.configurable_param(self.config.max)
    self.decay = self.gan.configurable_param(self.config.decay)
    allvars = self.gan.variables()
    if self.config.components:
        allvars = []
        for c in self.config.components:
            allvars += getattr(self.gan, c).variables()
    self.update_weight_constraints = [self._update_weight_constraint(v,i) for i,v in enumerate(allvars)]
    self.update_weight_constraints = [v for v in self.update_weight_constraints if v is not None]

  def _update_diversity(self, v, i):
    #Diversity regularized adversarial training
    if len(self.ops.shape(v)) == 0:
        return tf.no_op()
    w = tf.reshape(v, [-1, self.ops.shape(v)[-1]])
    wt = tf.transpose(w)
    wt_n = tf.math.l2_normalize(wt)
    s1 = tf.matmul(wt_n, tf.transpose(wt_n))
    print("W  : ", self.ops.shape(w))
    print("WT : ", self.ops.shape(wt))
    print("WTN: ", self.ops.shape(wt_n))
    print("S1 : ", self.ops.shape(s1), "wt dot wt_n")
    s2 = tf.reshape(s1, self.ops.shape(v))
    sim = s2 * (tf.zeros_like(s2)+tf.eye(self.ops.shape(s2)[0]))
    print("SIM : ", self.ops.shape(sim), "eye")
    mask = tf.reshape(sim * tf.nn.relu(tf.sign(tf.abs(sim))+1e-8), self.ops.shape(v))
    return tf.assign(v, mask)

  def _update_ortho(self,v,i):
    s = self.gan.ops.shape(v)
    if len(s) == 4 and s[0] == s[1]:
      w=v
      newv = []
      #s = self.ops.shape(v_transpose)
      #identity = tf.reshape(identity, [s[0],s[1],1,1])
      #identity = tf.tile(identity, [1,1,s[2],s[3]])
      w = tf.transpose(w, perm=[2,3,0,1])
      for i in range(self.config.iterations or 3):
          wt = tf.transpose(w, perm=[1,0,2,3])
          w2 = tf.reshape(w,[-1, s[0],s[1]])
          wt2 = tf.reshape(wt,[-1, s[0],s[1]])
          wtw = tf.matmul(wt2,w2)
          eye = tf.eye(s[0],s[1])
          eye = tf.tile(eye, [1,s[2]*s[3]])
          eye = tf.reshape(eye, self.gan.ops.shape(w))
          wtw = tf.reshape(wtw, self.gan.ops.shape(w))
          qk = eye - wtw
          w = w * (eye + 0.5*qk)
      w = tf.transpose(w, perm=[2,3,0,1])
      newv = w
      newv=(1.0+self.decay)*v - self.decay*(newv)
      newv = tf.reshape(newv,self.ops.shape(v))
      return tf.assign(v, newv)
    else:
      return None


  def _update_ortho2(self,v,i):
    if len(v.shape) == 4:
      w=v
      w = tf.transpose(w, perm=[2,3,0,1])
      identity = tf.cast(tf.diag(np.ones(self.ops.shape(w)[0])), tf.float32)
      wt = tf.transpose(w, perm=[1,0,2,3])
      #s = self.ops.shape(v_transpose)
      #identity = tf.reshape(identity, [s[0],s[1],1,1])
      #identity = tf.tile(identity, [1,1,s[2],s[3]])
      newv = tf.matmul(w, tf.matmul(wt,w))
      newv = tf.reshape(newv,self.ops.shape(v))
      newv = tf.transpose(newv, perm=[2,3,0,1])
      newv=(1+self.decay)*v - self.decay*(newv)

      return tf.assign(v, newv)
    return None
  def _update_lipschitz(self,v,i):
    config = self.config
    if len(v.shape) > 1:
      k = self.config.weight_constraint_k or 100.0000
      wi_hat = v
      if len(v.shape) == 4:
        #fij = tf.reduce_sum(tf.abs(wi_hat),  axis=[0,1])
        fij = wi_hat
        fij = tf.reduce_sum(tf.abs(fij),  axis=[1])
        fij = tf.reduce_max(fij,  axis=[0])
      else:
        fij = wi_hat

      if self.config.ortho_pnorm == "inf":
        wp = tf.reduce_max(tf.reduce_sum(tf.abs(fij), axis=0), axis=0)
      else:
        # conv
        wp = tf.reduce_max(tf.reduce_sum(tf.abs(fij), axis=1), axis=0)
      ratio = (1.0/tf.maximum(1.0, wp/k))
      
      if self.config.weight_bounce:
        bounce = tf.minimum(1.0, tf.ceil(wp/k-0.999))
        ratio -= tf.maximum(0.0, bounce) * 0.2

      if self.config.weight_scaleup:
        up = tf.minimum(1.0, tf.ceil(0.02-wp/k))
        ratio += tf.maximum(0.0, up) * k/wp * 0.2

      wi = ratio*(wi_hat)
      #self.gan.metrics['wi'+str(i)]=wp
      #self.gan.metrics['wk'+str(i)]=ratio
      #self.gan.metrics['bouce'+str(i)]=bounce
      return tf.assign(v, wi)
    return None

  def _update_l2nn(self,v,i):
    config = self.config
    s = self.gan.ops.shape(v)
    if len(v.shape) == 4 and s[0] == s[1]:
      w=v
      wt = tf.transpose(w, perm=[1,0,2,3])
      w2 = tf.reshape(w,[-1, s[0],s[1]])
      wt2 = tf.reshape(wt,[-1, s[0],s[1]])
      wtw = tf.matmul(wt2,w2)
      wwt = tf.matmul(w2,wt2)
      wtw = tf.reshape(wtw, [-1, self.ops.shape(v)[-1]])
      wwt = tf.reshape(wwt, [-1, self.ops.shape(v)[-1]])
    else:
      #w = v
      #w = tf.reshape(w, [-1, self.ops.shape(v)[-1]])
      #wt = tf.transpose(w)
      #wtw = tf.matmul(wt,w)
      #wwt = tf.matmul(w,wt)
      return None
    def _r(m):
      s = self.ops.shape(m)
      m = tf.abs(m)
      m = tf.reduce_sum(m, axis=0,keep_dims=True)
      m = tf.reduce_max(m, axis=1,keep_dims=True)
      #m = tf.tile(m,[s[0],s[1],1,1])
      return m
    bw = tf.minimum(_r(wtw), _r(wwt))
    #self.gan.add_metric('bw', tf.reduce_mean(bw))
    #wi = v-(tf.sign(v)*bw)#
    wi = (v/bw)
    if self.decay is not None:
      wi = (1-self.decay)*v+(self.decay*wi)
    wi = tf.reshape(wi, self.ops.shape(v))
    return tf.assign(v, wi)

  def _update_weight_constraint(self,v,i):
    if "Adam" in v.name or "AMSGrad" in v.name or "RMS" in v.name or "Adadelta" in v.name:
      print("> skipping(name)", v.name)
      return None

    config = self.config
    #skipped = [gan.generator.ops.weights[0], gan.generator.ops.weights[-1], gan.discriminator.ops.weights[0], gan.discriminator.ops.weights[-1]]
    #skipped = [gan.discriminator.ops.weights[-1]]
    skipped=[]
    for skip in skipped:
      if self.ops.shape(v) == self.ops.shape(skip):
        print("Skipping constraints on", v)
        return None
    constraints = self.config.constraints or self.config.weight_constraint or []
    result = []
    if "ortho" in constraints:
      result.append(self._update_ortho(v,i))
    if "ortho2" in constraints:
      result.append(self._update_ortho2(v,i))
    if "lipschitz" in constraints:
      result.append(self._update_lipschitz(v,i))
    if "l2nn" in constraints:
      result.append(self._update_l2nn(v,i))
    if "l2nn-d" in constraints:
      if v in d_vars:
        result.append(self._update_l2nn(v,i))
    if "diversity" in constraints:
        result.append(self._update_diversity(v, i))
    result = [r for r in result if r is not None]
    if(len(result) == 0):
      return None
    return result

  def before_step(self, step, feed_dict):
    if self.config.order == "after":
        pass
    else:
        if ((step % (self.config.constraint_every or 100)) == 0):
            #print("Applying weight constraint (pre)")
            self.gan.session.run(self.update_weight_constraints, feed_dict)

  def after_step(self, step, feed_dict):
    if self.config.order == "after":
        if ((step % (self.config.constraint_every or 100)) == 0):
            #print("Applying weight constraint (post)")
            self.gan.session.run(self.update_weight_constraints, feed_dict)



#From https://gist.github.com/EndingCredits/b5f35e84df10d46cfa716178d9c862a3
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from tensorflow.python.ops import control_flow_ops
from tensorflow.python.ops import math_ops
from tensorflow.python.ops import state_ops
from tensorflow.python.framework import ops
from tensorflow.python.training import optimizer
import hyperchamber as hc
import inspect

class GANOptimizer(optimizer.Optimizer):
  """
    Splits optimization targets between g_optimizer and g_optimizer
  """

  def __init__(self, _, gan=None, config=None, g_optimizer=None, d_optimizer=None, name="GANOptimizer"):
    super().__init__(config.learn_rate, name=name)
    self.gan = gan
    self.config = config

    self.d_optimizer = self.gan.create_optimizer(d_optimizer)
    self.g_optimizer = self.gan.create_optimizer(g_optimizer)

  def _prepare(self):
    super()._prepare()
    self.d_optimizer._prepare()
    self.g_optimizer._prepare()

  def variables(self):
    return self.d_optimizer.variables() + self.g_optimizer.variables()

  def get_slot_names(self):
    return self.d_optimizer.get_slot_names() + self.g_optimizer.get_slotnames()

  def _create_slots(self, var_list):
    super()._create_slots(var_list)
    d_vars = [v for v in var_list if v in self.gan.d_vars()]
    g_vars = [v for v in var_list if v in self.gan.g_vars()]
    self.d_optimizer._create_slots(d_vars)
    self.g_optimizer._create_slots(g_vars)
    missing_vars = [v for v in var_list if v not in self.gan.g_vars() + self.gan.d_vars()]
    if len(missing_vars) > 0:
        print("Error, GANOptimizer does not know how to handle missing variables (not in d_vars or g_vars)", missing_vars)
        raise("Error, GANOptimizer does not know how to handle missing variables (not in d_vars or g_vars)")

  def _apply_dense(self, grad, var):
    if var in self.gan.d_vars():
        return self.d_optimizer._apply_dense(grad, var)
    elif var in self.gan.g_vars():
        return self.g_optimizer._apply_dense(grad, var)
    raise("Unable to handle", var)


  def get_slot_names(self):
      return list(set(self.d_optimizer.get_slot_names() + self.g_optimizer.get_slot_names()))

  def get_slot(self, var, name):
      return self.d_optimizer.get_slot(var, name) or self.g_optimizer.get_slot(var, name)
  
  def _apply_sparse(self, grad, var):
    raise NotImplementedError("Sparse gradient updates are not supported.")

  def apply_gradients(self, grads_and_vars, global_step=None, name=None):
    d_vars = self.gan.d_vars()
    args = []
    if len(grads_and_vars) > len(d_vars):
        args += [self.g_optimizer.apply_gradients(grads_and_vars[len(d_vars):], global_step, name)]
    args += [self.d_optimizer.apply_gradients(grads_and_vars[:len(d_vars)], global_step, name)]
    return tf.group(*args)



import hyperchamber as hc
import inspect
import copy
import os
import operator
from functools import reduce

from .base_discriminator import BaseDiscriminator
from hypergan.configurable_component import ConfigurableComponent

class ConfigurableDiscriminator(ConfigurableComponent):
    def __init__(self, gan, config, *args, **kw_args):
        ConfigurableComponent.__init__(self, gan, config,*args, **kw_args)




import importlib
import json
import numpy as np
import os
import re
import sys
import time
import uuid
import copy

from hypergan.discriminators import *
from hypergan.distributions import *
from hypergan.generators import *
from hypergan.inputs import *
from hypergan.samplers import *
from hypergan.trainers import *

import hyperchamber as hc
from hyperchamber import Config
from hypergan.ops import TensorflowOps
import hypergan as hg

from hypergan.gan_component import ValidationException, GANComponent
from .base_gan import BaseGAN

class ConfigurableGAN(BaseGAN):
    def __init__(self, *args, **kwargs):
        self.d_terms = []
        self.Ds = []
        BaseGAN.__init__(self, *args, **kwargs)

    def create_encoder(self):
        return self.create_component(self.config.encoder)

    def create_latent(self, zi):
        return self.create_component(self.config.latent)

    def create_generator(self):
        return self.create_component(self.config.generator)

    def parse_opts(self, opts):
        options = {}
        for opt in opts.split(","):
            if opt == "":
                continue
            name, val = opt.split("=")
            value = self.configurable_param(val)
            options[name]=value
        return hc.Config(options)

    def required(self):
        return "terms".split()

    def create_term(self, term):
        for match in 
        matching = {
                "gN(eN(xN))": self.create_generator,
                "gN(zN)": self.create_generator
                }
        for regex, method in matching.items():
            regex_subbed = regex.replace("(", '\(').replace(")", '\)').replace("N", "(\d+)?").replace(",options", "([-,=\w\d\.\(\)]+)?")
            regex = re.compile(regex_subbed)
            args = re.match(regex, term)
            if args:
                return method(*args.groups())

        raise ValidationException("Could not match term: " + term)

    def forward_term(self, term):
        matching = {
                "gN(eN(xN))": self.geN,
                "gN(zN)": self.gzN,
                "xN": self.xN
                }
        for regex, method in matching.items():
            regex_subbed = regex.replace("(", '\(').replace(")", '\)').replace("N", "(\d+)?").replace(",options", "([-,=\w\d\.\(\)]+)?")
            regex = re.compile(regex_subbed)
            args = re.match(regex, term)
            if args:
                return method(*args.groups())

        raise ValidationException("Could not match term: " + term)

    def create(self):
        config = self.config

        self.latent = hc.Config({"sample": self.zN(0)})
        self.discriminators = []
        self.losses = []
        for i,term in enumerate(self.config.terms):
            dN, args = term.split(":")

            d_terms = args.split("/")
            terms = []
            for dt in d_terms:
                terms += (term,self.create_term(dt))
            reuse = False

            dN = re.findall("\d+", dN)[0]
            dN = int(dN)
            tfname = "d"+str(dN)
            D = self.create_component(config.discriminator)
            self.Ds.append(D)
            self.d_terms += terms
 
        self.trainer = self.create_component(config.trainer)

    def create_controls(self, z_shape):
        direction = tf.constant(0.0, shape=z_shape, name='direction') * 1.00
        slider = tf.constant(0.0, name='slider', dtype=tf.float32) * 1.00
        return direction, slider

    def forward_pass(self):
        d_reals = []
        d_fakes = []
        for terms in d_terms:

        return d_reals, d_fakes

    def forward_loss(self):
        losses = []
        for d_real, d_fake in zip(d_reals, d_fakes):
            loss = self.create_component(config.loss, discriminator=d, split=len(d_terms))
            d_loss, g_loss = loss.forward(d_real, d_fake)
            d_loss = [self.configurable_param(config.term_gammas[i]) * d_loss, self.configurable_param(config.term_gammas[i]) * g_loss]
            losses += [[d_loss, g_loss]]

        self.loss = hc.Config({
            'sample': [sum([l.sample[0] for l in losses]), sum([l.sample[1] for l in losses])]
            })

    def g_parameters(self):
        params = []
        for d_terms in self.d_terms:
            for term in d_terms:
                params += term[1].parameters()
        return params

    def d_parameters(self):
        params = []
        for m in self.Ds:
            params += m.parameters()
        return params



from hypergan.gan_component import GANComponent
import numpy as np
import torch

class BaseLoss(GANComponent):
    def __init__(self, gan, config):
        super(BaseLoss, self).__init__(gan, config)
        self.relu = torch.nn.ReLU()

    def create(self, *args):
        pass

    def required(self):
        return "".split()

    def forward(self, d_real, d_fake):
        d_loss, g_loss = [c.mean() for c in self._forward(d_real, d_fake)]

        return d_loss, g_loss

    def forward_adversarial_norm(self, d_real, d_fake):
        return (torch.sign(d_real-d_fake)*((d_real - d_fake)**2)).mean()
        #return 0.5 * (self.dist(d_real,d_fake) + self.dist(d_fake, d_real)).sum()



import hyperchamber as hc

from hypergan.losses.base_loss import BaseLoss

class LeastSquaresLoss(BaseLoss):

    def required(self):
        return "".split()

    def _forward(self, d_real, d_fake):
        config = self.config

        a,b,c = (config.labels or [-1,1,1])
        d_loss = 0.5*((d_real - b)**2) + 0.5*((d_fake - a)**2)
        g_loss = 0.5*((d_fake - c)**2)

        return [d_loss, g_loss]



import math
import torch
import torch.nn as nn

class MultiHeadAttention(nn.Module):
    def __init__(self, input_size, output_size, heads):
        super(MultiHeadAttention,self).__init__()
        self.heads = heads
        self.features = input_size // heads
        self.features_sqrt = torch.tensor(math.sqrt(self.features)).cuda()

        self.f = nn.Linear(input_size, input_size)
        self.g = nn.Linear(input_size, input_size)
        self.h = nn.Linear(input_size, input_size)
        self.o = nn.Linear(input_size, output_size)

        self.softmax  = nn.Softmax(dim=-1)
    def forward(self,x):
        batch_size, input_size = x.shape
        f  = self.f(x).view(batch_size,1,self.heads,self.features).permute(0,2,3,1)
        g =  self.g(x).view(batch_size,1,self.heads,self.features).permute(0,2,1,3)
        fg =  torch.matmul(g, f) / self.features_sqrt
        attention_map = self.softmax(fg)
        h = self.h(x).view(batch_size,1,self.heads,self.features).permute(0,2,1,3)

        fgh = torch.matmul(attention_map, h)
        output = fgh.permute(0,2,1,3).view(x.shape)
        return self.o(output)



import numpy as np
import torch
import hyperchamber as hc
import inspect

from hypergan.gan_component import ValidationException, GANComponent
from hypergan.trainers.base_trainer import BaseTrainer
from hypergan.optimizers.adamirror import Adamirror

TINY = 1e-12

class SimultaneousTrainer(BaseTrainer):
    """ Steps G and D simultaneously """
    def _create(self):
        self.optimizer = self.create_optimizer()

    def required(self):
        return "optimizer".split()

    def _step(self, feed_dict):
        metrics = self.gan.metrics()

        self.before_step(self.current_step, feed_dict)
        d_grads, g_grads = self.calculate_gradients()

        for hook in self.train_hooks:
            d_grads, g_grads = hook.gradients(d_grads, g_grads)
        for p, np in zip(self.gan.d_parameters(), d_grads):
            p.grad = np
        for p, np in zip(self.gan.g_parameters(), g_grads):
            p.grad = np

        self.optimizer.step()

        if self.current_step % 10 == 0:
            self.print_metrics(self.current_step)

    def calculate_gradients(self):
        self.optimizer.zero_grad()
        d_loss, g_loss = self.gan.forward_loss()
        self.gan.add_metric('d_loss', d_loss.mean())
        self.gan.add_metric('g_loss', g_loss.mean())
        for hook in self.train_hooks:
            loss = hook.forward(d_loss, g_loss)
            if loss[0] is not None:
                d_loss += loss[0]
            if loss[1] is not None:
                g_loss += loss[1]

        self.gan.set_generator_trainable(True)
        self.gan.set_discriminator_trainable(False)

        g_loss.mean().backward(retain_graph=True)

        self.gan.set_generator_trainable(False)
        self.gan.set_discriminator_trainable(True)

        d_loss.mean().backward(retain_graph=True)
        self.gan.set_generator_trainable(True)

        d_grads = [p.grad for p in self.gan.d_parameters()]
        g_grads = [p.grad for p in self.gan.g_parameters()]
        return d_grads, g_grads

    def print_metrics(self, step):
        metrics = self.gan.metrics()
        metric_values = self.output_variables(metrics)
        print(str(self.output_string(metrics) % tuple([step] + metric_values)))




import numpy as np
import hyperchamber as hc
import inspect

from hypergan.trainers.base_trainer import BaseTrainer

TINY = 1e-12

class ConsensusTrainer(BaseTrainer):
    def create(self):
        config = self.config
        lr = config.learn_rate
        self.global_step = tf.train.get_global_step()
        decay_function = config.decay_function
        if decay_function:
            print("!!using decay function", decay_function)
            decay_steps = config.decay_steps or 50000
            decay_rate = config.decay_rate or 0.9
            decay_staircase = config.decay_staircase or False
            self.lr = decay_function(lr, self.global_step, decay_steps, decay_rate, decay_staircase)
        else:
            self.lr = lr

        return self._create()


    def _create(self):
        gan = self.gan
        config = self.config

        d_vars = self.d_vars or gan.discriminator.variables()
        g_vars = self.g_vars or (gan.encoder.variables() + gan.generator.variables())

        loss = self.loss or gan.loss
        d_loss, g_loss = loss.sample
        allloss = d_loss + g_loss
        allvars = d_vars + g_vars

        d_grads = tf.gradients(d_loss, d_vars)
        g_grads = tf.gradients(g_loss, g_vars)

        grads = d_grads + g_grads

        self.d_log = -tf.log(tf.abs(d_loss+TINY))
        for g, d_v in zip(grads,d_vars):
            if g is None:
                print("!!missing gradient")
                print(d_v)
                return
        reg = 0.5 * sum(
            tf.reduce_sum(tf.square(g)) for g in grads if g is not None
        )
        # Jacobian times gradiant
        if config.update_rule == "ttur" or config.update_rule == 'single-step':
            Jgrads = [0 for i in allvars]
        else:
            Jgrads = tf.gradients(reg, allvars)

        print("JG", Jgrads)

        self.g_gradient = tf.ones([1])
        def amp_for(v):
            if v in g_vars:
                return config.g_w_lambda or 3
            if v in d_vars:
                return config.d_w_lambda or 1

        def applyvec(g, jg, v, decay):
            prev = v
            nextw = v+g + jg * (config.jg_alpha or 0.1)
            if decay is not None:
                return ((decay) * prev + (1.0-decay)*nextw)-v
            else:
                return nextw-v

        def gradient_for(g, jg, v, decay):
            if config.update_rule == "ttur":
                if decay is not None:
                    amp = v+amp_for(v)*g
                    ng = ((decay) * v + (1.0-decay)*amp)-v
                else:
                    ng = amp_for(v)*g
            else:
                if decay is not None:
                    if v in g_vars:
                        ng = applyvec(g, jg, v, decay)
                    else:
                        ng = applyvec(g, jg, v, None)
                else:
                    ng = applyvec(g, jg, v, decay)
            return ng
        decay = config.g_exponential_moving_average_decay
        apply_vec = [ (gradient_for(g, Jg, v, decay), v) for (g, Jg, v) in zip(grads, Jgrads, allvars) if Jg is not None ]
        apply_vec_d = [ (gradient_for(g, Jg, v, decay), v) for (g, Jg, v) in zip(d_grads, Jgrads[:len(d_vars)], d_vars) if Jg is not None ]
        apply_vec_g = [ (gradient_for(g, Jg, v, decay), v) for (g, Jg, v) in zip(g_grads, Jgrads[len(d_vars):], g_vars) if Jg is not None ]

        defn = {k: v for k, v in config.items() if k in inspect.getargspec(config.trainer).args}
        tr = config.trainer(self.lr, **defn)


        optimizer = tr.apply_gradients(apply_vec, global_step=self.global_step)
        d_optimizer = tr.apply_gradients(apply_vec_d, global_step=self.global_step)
        g_optimizer = tr.apply_gradients(apply_vec_g, global_step=self.global_step)

        self.g_loss = g_loss
        self.d_loss = d_loss
        self.optimizer = optimizer
        self.d_optimizer = d_optimizer
        self.g_optimizer = g_optimizer

        return optimizer, optimizer

    def required(self):
        return "trainer learn_rate".split()

    def _step(self, feed_dict):
        gan = self.gan
        sess = gan.session
        config = self.config
        loss = self.loss or gan.loss
        metrics = gan.metrics()

        metric_values = sess.run([self.optimizer] + self.output_variables(metrics), feed_dict)[1:]

        if self.current_step % 100 == 0:
            print(self.output_string(metrics) % tuple([self.current_step] + metric_values))




import torch
import hyperchamber as hc
import numpy as np
import inspect
from operator import itemgetter
from hypergan.train_hooks.base_train_hook import BaseTrainHook
from torch.nn.parameter import Parameter
from torch.autograd import Variable
from torch.autograd import grad as torch_grad

class AdversarialNormTrainHook(BaseTrainHook):
    def __init__(self, gan=None, config=None, trainer=None):
        super().__init__(config=config, gan=gan, trainer=trainer)
        self.d_loss = None
        self.g_loss = None
        if self.config.gamma is not None:
            self.gamma = torch.Tensor([self.config.gamma]).float()[0].cuda()#self.gan.configurable_param(self.config.gamma or 1.0)
        if self.config.gammas is not None:
            self.gammas = [
                        torch.Tensor([self.config.gammas[0]]).float()[0].cuda(),#self.gan.configurable_param(self.config.gamma or 1.0)
                        torch.Tensor([self.config.gammas[1]]).float()[0].cuda()#self.gan.configurable_param(self.config.gamma or 1.0)
                    ]
        self.relu = torch.nn.ReLU()
        self.target = [Parameter(x, requires_grad=True) for x in self.gan.discriminator_real_inputs()]
        self.x_mod_target = torch.zeros_like(self.target[0])
        self.g_mod_target = torch.zeros_like(self.target[0])

    def forward(self, d_loss, g_loss):
        if self.config.mode == "real" or self.config.mode is None:
            for target, data in zip(self.target, self.gan.discriminator_real_inputs()):
                target.data = data.clone()
            d_fake = self.gan.d_fake
            d_real = self.gan.forward_discriminator(self.target)
            loss, _, mod_target = self.regularize_adversarial_norm(d_fake, d_real, self.target)
            norm = (-((mod_target[0] - self.gan.discriminator_real_inputs()[0])**2)).mean()
            for mt, t in zip(mod_target[1:], self.gan.discriminator_real_inputs()[1:]):
                norm += (-((mt - t) ** 2)).mean()
            if self.config.forward_discriminator:
                dadv = self.gan.forward_discriminator(mod_target)
                norm += (-((dadv - self.gan.d_real) ** 2)).mean()
        elif self.config.mode == "fake":
            for target, data in zip(self.target, self.gan.discriminator_fake_inputs()):
                target.data = data.clone()
            d_fake = self.gan.forward_discriminator(self.target)
            d_real = self.gan.d_real
            loss, norm, mod_target = self.regularize_adversarial_norm(d_real, d_fake, self.target)
            norm = (-((mod_target[0] - self.gan.discriminator_fake_inputs()[0])**2)).mean()
            for mt, t in zip(mod_target[1:], self.gan.discriminator_fake_inputs()[1:]):
                norm += (-((mt - t) ** 2)).mean()
            if self.config.forward_discriminator:
                dadv = self.gan.forward_discriminator(mod_target)
                norm += (-((dadv - self.gan.d_fake) ** 2)).mean()

        if self.config.loss:
          if "g" in self.config.loss:
              self.g_loss = self.gamma * norm.mean()
              self.gan.add_metric('an_g', self.g_loss)
          if "d" in self.config.loss:
              self.d_loss = self.gamma * norm.mean()
              self.gan.add_metric('an_d', self.d_loss)
          if "dg" in self.config.loss:
              self.d_loss = self.gammas[0] * norm.mean()
              self.gan.add_metric('an_d', self.d_loss)
              self.g_loss = self.gammas[1] * norm.mean()
              self.gan.add_metric('an_g', self.g_loss)
        else:
            self.d_loss = self.gamma * norm.mean()
            self.gan.add_metric('an_d', self.d_loss)

        return [self.d_loss, self.g_loss]

    def regularize_adversarial_norm(self, d1_logits, d2_logits, target):
        loss = self.gan.loss.forward_adversarial_norm(d1_logits, d2_logits)

        d1_grads = torch_grad(outputs=loss, inputs=target, retain_graph=True, create_graph=True)
        mod_target = [_d1 + _t for _d1, _t in zip(d1_grads, target)]

        return loss, None, mod_target



#From https://gist.github.com/EndingCredits/b5f35e84df10d46cfa716178d9c862a3
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from tensorflow.python.ops import control_flow_ops
from tensorflow.python.ops import math_ops
from tensorflow.python.ops import state_ops
from tensorflow.python.framework import ops
from tensorflow.python.training import optimizer
import hyperchamber as hc
import numpy as np
import inspect
from operator import itemgetter
from hypergan.train_hooks.base_train_hook import BaseTrainHook

class MinibatchTrainHook(BaseTrainHook):
  def __init__(self, gan=None, config=None, trainer=None, name="MinibatchTrainHook"):
    super().__init__(config=config, gan=gan, trainer=trainer, name=name)

    config = self.config
    setattr(gan,config.name or 'minibatch', self)

    net = gan.discriminator.layer(config.target or 'minibatch')
    minibatch_size = gan.batch_size()*2
    single_batch_size = gan.batch_size()
    if self.config.gan_input is None:
        n_kernels = config.kernels or 256
        dim_per_kernel = config.dim_per_kernel or 10
        print("[discriminator] minibatch from", net, "to", n_kernels*dim_per_kernel)
        net = tf.reshape(net, [minibatch_size, -1])
        initializer = self.ops.lookup_initializer("stylegan", config.initializer_options or {})
        x = self.ops.linear(net, n_kernels * dim_per_kernel, initializer=initializer)
        activation = tf.reshape(x, (minibatch_size, n_kernels, dim_per_kernel))

        big = np.zeros((minibatch_size, minibatch_size))
        big += np.eye(minibatch_size)
        big = tf.expand_dims(big, 1)
        big = tf.cast(big,dtype=tf.float32)

        abs_dif = tf.reduce_sum(tf.abs(tf.expand_dims(activation,3) - tf.expand_dims(tf.transpose(activation, [1, 2, 0]), 0)), 2)
        mask = 1. - big
        masked = tf.exp(-abs_dif) * mask
        def half(tens, second):
            m, n, _ = tens.get_shape()
            m = int(m)
            n = int(n)
            return tf.slice(tens, [0, 0, second * single_batch_size], [m, n, single_batch_size])

        f1 = tf.reduce_sum(half(masked, 0), 2) / tf.reduce_sum(half(mask, 0))
        f2 = tf.reduce_sum(half(masked, 1), 2) / tf.reduce_sum(half(mask, 1))
        stack_f = tf.concat([f1,f2], axis=0)

    if self.config.type == 'gan':
        if self.config.gan_input == 'distance':
            real, fake = tf.split(net, 2, 0)
            stack_f = tf.concat([real - fake,fake-real],  axis=0)
        if self.config.gan_input == 'moments':
            real, fake = tf.split(net, 2, 0)
            rm,rv = tf.nn.moments(real, 0)
            fm,fv = tf.nn.moments(fake, 0)
            rm = tf.expand_dims(rm,0)
            rv = tf.expand_dims(rv,0)
            fm = tf.expand_dims(fm,0)
            fv = tf.expand_dims(fv,0)
            print("_____", fm)
            r = tf.concat([rm, rv], axis=0)
            f = tf.concat([fm, fv], axis=0)
            stack_f = tf.concat([r,f],  axis=0)
        minibatch_discriminator = gan.create_component(self.config.discriminator or gan.config.minibatch_discriminator, name=(config.name or 'minibatch_discriminator'), input=stack_f)
        setattr(gan, config.name+"_discriminator", minibatch_discriminator)
        gan.discriminator.add_variables(minibatch_discriminator)
        loss = gan.create_component(self.gan.config.loss, minibatch_discriminator)
        gan.add_metric('mini_dl', loss.sample[0])
        gan.add_metric('mini_gl', loss.sample[1])
        self.loss = loss.sample
    else:
        self.minibatch_value = tf.concat([f1, f2], axis=1)
        self.minibatch_feed = gan.discriminator.layer('minibatch_feed')
        self.loss = [None, None]
    gan.discriminator.add_variables(self)

  def losses(self):
      return self.loss

  def after_step(self, step, feed_dict):
      pass

  def before_step(self, step, feed_dict):
    if self.config.type != 'gan':
      feed_dict[self.minibatch_feed]=self.gan.session.run(self.minibatch_value)

    pass



from .base_gan import BaseGAN
from hyperchamber import Config
from hypergan.discriminators import *
from hypergan.distributions import *
from hypergan.gan_component import ValidationException, GANComponent
from hypergan.generators import *
from hypergan.inputs import *
from hypergan.samplers import *
from hypergan.trainers import *
import copy
import hyperchamber as hc
import hypergan as hg
import importlib
import json
import numpy as np
import os
import sys
import time
import torch
import uuid


class EncoderGAN(BaseGAN):
    """ 
    """
    def __init__(self, *args, **kwargs):
        BaseGAN.__init__(self, *args, **kwargs)

    def create(self):
        self.latent = self.create_component("latent")
        self.x = self.inputs.next()[0]
        self.z = self.latent.next()
        self.generator = self.create_component("generator", input=self.latent)
        self.encoder = self.create_component("encoder", input=self.x)
        self.discriminator = self.create_component("discriminator")
        self.discriminator2 = self.create_component("z_discriminator", input=self.latent)
        self.loss = self.create_component("loss")
        self.trainer = self.create_component("trainer")

    def forward_discriminator(self, inputs):
        return self.discriminator(inputs[0]) + self.discriminator2(inputs[1])

    def forward_pass(self):
        self.x = self.inputs.next()
        self.z = self.latent.next()
        g = self.generator(self.z)
        e = self.encoder(g)
        self.g = g
        self.e = e
        d_real = self.forward_discriminator([self.x, self.z])
        d_fake = self.forward_discriminator([self.g, self.e])
        self.d_fake = d_fake
        self.d_real = d_real
        return d_real, d_fake

    def discriminator_components(self):
        return [self.discriminator, self.discriminator2]

    def generator_components(self):
        return [self.generator, self.encoder]

    def discriminator_fake_inputs(self, discriminator_index=0):
        return [self.g, self.e]

    def discriminator_real_inputs(self, discriminator_index=0):
        if hasattr(self, 'x'):
            return [self.x, self.z]
        else:
            return [self.inputs.next(), self.latent.next()]



from hypergan.gan_component import GANComponent

class BaseDiscriminator(GANComponent):
    def __init__(self, gan, config, name=None, input=None, reuse=None, features=None, weights=None, biases=None):
        GANComponent.__init__(self, gan, config)
        self.input = input
        self.name = name
        self.features = features

    def create(self, net=None):
        pass



#From https://gist.github.com/EndingCredits/b5f35e84df10d46cfa716178d9c862a3
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from tensorflow.python.ops import control_flow_ops
from tensorflow.python.ops import math_ops
from tensorflow.python.ops import state_ops
from tensorflow.python.framework import ops
from tensorflow.python.training import optimizer
import hyperchamber as hc
import numpy as np
import inspect
from operator import itemgetter
from hypergan.train_hooks.base_train_hook import BaseTrainHook

class AdversarialRobustTrainHook(BaseTrainHook):
  "AR from https://openreview.net/pdf?id=HJE6X305Fm"
  def __init__(self, gan=None, config=None, trainer=None, name="GradientPenaltyTrainHook"):
    super().__init__(config=config, gan=gan, trainer=trainer, name=name)
    if hasattr(self.gan, 'x0'):
        gan_inputs = self.gan.x0
    else:
        gan_inputs = self.gan.inputs.x
    config = hc.Config(config)

    if 'lambda' in config:
        self._lambda = self.gan.configurable_param(config['lambda'])
    else:
        self._lambda = 1.0
    self._vlambda = self.gan.configurable_param(config.vlambda or 1.0)

    self.trainablex=tf.Variable(tf.zeros_like(self.gan.inputs.x))
    self.trainableg=tf.Variable(tf.zeros_like(self.gan.inputs.x))
    clearx=tf.assign(self.trainablex, tf.zeros_like(self.gan.inputs.x))
    clearg=tf.assign(self.trainableg, tf.zeros_like(self.gan.inputs.x))
    ops = self.gan.ops
    with tf.get_default_graph().control_dependencies([clearx, clearg]):
        self.adversarial_discriminator = gan.create_component(gan.config.discriminator, name="discriminator", input=tf.concat([gan.inputs.x+self.trainablex,gan.generator.sample+self.trainableg],axis=0), features=[gan.features], reuse=True)
        self.v = tf.gradients(self.adversarial_discriminator.sample, [self.trainablex, self.trainableg])
        self.v = [self._vlambda*v/tf.norm(v, ord=2) for v in self.v]

        self.robustness_discriminator = gan.create_component(gan.config.discriminator, name='discriminator', input=tf.concat([gan.inputs.x+self.v[0], gan.generator.sample+self.v[1]], axis=0), features=[gan.features], reuse=True)

        if config.loss_type == 'gan':
            self.loss = gan.create_component(gan.config.loss, self.robustness_discriminator).sample
            self.loss = [self._lambda * self.loss[0], self._lambda * self.loss[1]]
        elif config.loss_type == 'valp':
            # https://arxiv.org/abs/1907.05681
            dy = tf.abs(self.robustness_discriminator.sample - gan.discriminator.sample)
            dy = tf.split(dy, 2)
            dx = [tf.norm(tf.reshape(self.v[0], [self.ops.shape(self.v[0])[0], -1]), axis=1), tf.norm(tf.reshape(self.v[1], [self.ops.shape(self.v[1])[0], -1]), axis=1)]
            dx = [tf.reshape(dx[0], [self.ops.shape(dx[0])[0], 1]), tf.reshape(dx[1], [self.ops.shape(dx[1])[0], 1])]
            self.gan.add_metric('dxlp', ops.squash(dx[0]))
            self.gan.add_metric('dylp', ops.squash(dy[0]))
            print('dx', dx, dy, dy[0]/dx[0])
            lvalp = [(dy[0]/(dx[0]+1e-32) - (self.config.k or 1.0)), (dy[1]/(dx[1]+1e-32) - (self.config.k or 1.0))]
            lvalp = [self.ops.squash(lvalp[0], self.gan.loss.config.reduce or tf.reduce_mean), self.ops.squash(lvalp[1], self.gan.loss.config.reduce or tf.reduce_mean)]
            #self.gan.add_metric('lvalp', lvalp[0])

            lx = self._lambda / 2.0 * tf.square(lvalp[0])
            if config.vlambda1 != 0:
              lx += -self._lambda * lvalp[0] 
            lg = self._lambda / 2.0 * tf.square(lvalp[1])
            if config.vlambda1 != 0:
              lg += -self._lambda * lvalp[1] 
            self.loss = [lx+lg, lg]
        else:
            self.loss = tf.square(self.robustness_discriminator.sample - gan.discriminator.sample)
            self.loss = [self._lambda * self.loss[0], self._lambda * self.loss[1]]


        #self.loss = gan.create_component(gan.config.loss, self.robustness_discriminator)
        self.gan.add_metric('adl', ops.squash(self.loss[0]))
        if self.loss[1] is not None:
            self.gan.add_metric('agl', ops.squash(self.loss[1]))
        #self.gan.add_metric('vx', tf.reduce_sum(self.v[0]))
        #self.gan.add_metric('vg', tf.reduce_sum(self.v[1]))

  def variables(self):
      return [self.trainablex, self.trainableg]
  def losses(self):
    return self.loss



import numpy as np
import hyperchamber as hc
import inspect

from hypergan.trainers.base_trainer import BaseTrainer

TINY = 1e-12

class DepthTrainer(BaseTrainer):
    """ Runs an optimizer multiple times and combines the output into a mixture. """
    def _create(self):
        self.hist = [0 for i in range(2)]
        config = self.config
        self.mix_threshold_reached = False
        variables = self.gan.d_vars() + self.gan.g_vars()
        self.ema = [ tf.Variable(_v) for _v in variables ]
        self.store_v = [ _v.assign(_v2) for _v,_v2 in zip(self.ema, variables) ]
        self.combine = [ _v.assign((config.decay or 0.1) *_ema + (1.-(config.decay or 0.1))*_new) for _v, _ema, _new in zip(variables, self.ema, variables)]
        self._delegate = self.gan.create_component(config.trainer, d_vars=self.d_vars, g_vars=self.g_vars)
        self.reset_optimizer_t = tf.variables_initializer(self._delegate.variables())
        self.depth_step = 0
        self.fitness = -self.gan.loss.d_fake
        self.latent = None

    def required(self):
        return "".split()

    def _best_latent(self):
        if self.latent is None:
            self.latent = self.gan.session.run(self.gan.latent.sample)
        fitness = self.gan.session.run(self.fitness, {self.gan.latent.sample:self.latent})
        zs = self.latent
        sort_zs = None
        last_fitness = 10000
        count = 0
        while True:
            d = self.gan.session.run([self.fitness,self.gan.latent.sample])
            _f = d[0]
            _z = d[1]
            fitness = np.reshape(fitness, np.shape(_f))
            fitness = np.concatenate([fitness,_f], axis=0)
            zs = np.reshape(zs, np.shape(_z))
            zs = np.concatenate([zs,_z], axis=0)
            sort = np.argsort(fitness.flatten())[:self.gan.batch_size()]
            zs = zs[sort]
            fitness = fitness.flatten()[sort]
            if fitness.flatten()[-1] < last_fitness:
                last_fitness = fitness[-1]
                count = 0
            else:
                count += 1
                if count > self.config.heuristic:
                    #print("z fit ", i)
                    sort_zs = np.reshape(zs, np.shape(_z))
                    break
        return sort_zs

    def _step(self, feed_dict):
        gan = self.gan
        sess = gan.session
        config = self.config
        depth = self.config.depth
        if depth:
            if self.current_step % depth == 0:
                if self.config.freeze_latent:
                    if self.config.freeze_latent == "best":
                        self.latent = self._best_latent()
                    else:
                        self.latent = self.gan.session.run(self.gan.latent.sample)
                    feed_dict[gan.latent.sample] = self.latent
                self.before_step(self.current_step, feed_dict)
                gan.session.run(self.store_v)
                if self.config.reset_optimizer:
                    self.gan.session.run([self.reset_optimizer_t])

            if self.config.freeze_latent:
                feed_dict[gan.latent.sample] = self.latent
            self._delegate.step(feed_dict)
            if self.current_step % depth == depth - 1:
                gan.session.run(self.combine)
                self.after_step(self.current_step, feed_dict)
        else:
            if self.depth_step == 0:
                if self.config.freeze_latent:
                    if self.config.freeze_latent == "best":
                        self.latent = self._best_latent()
                    else:
                        self.latent = self.gan.session.run(self.gan.latent.sample)
                    feed_dict[gan.latent.sample] = self.latent
                self.before_step(self.current_step, feed_dict)
                gan.session.run(self.store_v)
                self.max_gradient_mean = 0.0
            if self.config.freeze_latent:
                feed_dict[gan.latent.sample] = self.latent
            self._delegate.step(feed_dict)
            gradient_mean = gan.session.run(gan.gradient_mean, feed_dict)
            self.depth_step += 1
            if gradient_mean > self.max_gradient_mean:
                self.max_gradient_mean = gradient_mean
            if gradient_mean/self.max_gradient_mean < (0.2 or self.config.gradient_threshold):
                gan.session.run(self.combine)
                self.after_step(self.current_step, feed_dict)
                self.depth_step = 0

    def variables(self):
        return self._delegate.variables()



import torch.nn as nn
import torch

class LearnedNoise(nn.Module):
    def __init__(self, batch_size, c, h, w, mul=0.1):
        super().__init__()
        self.batch_size = batch_size
        self.h = h
        self.w = w
        self.c = c
        self.weight = nn.Parameter(torch.randn(1, self.c, 1, 1) * mul)
        self.noise = torch.Tensor(self.batch_size, 1, self.h, self.w).cuda()

    def forward(self, input):
        self.noise.normal_()
        return self.noise * self.weight



import PIL
import torch.nn.functional as F
import torch
from collections.abc import Sequence, Iterable
import warnings

from torchvision.transforms import functional as F


class CropResizeTransform(object):
    """Crop the image to min(h,w) then resize it

    Args:
        size (sequence or int): Desired output size. If size is a sequence like
            (h, w), output size will be matched to this. If size is an int,
            smaller edge of the image will be matched to this number.
            i.e, if height > width, then image will be rescaled to
            (size * height / width, size)
        interpolation (int, optional): Desired interpolation. Default is
            ``PIL.Image.BILINEAR``
    """

    def __init__(self, size, interpolation=PIL.Image.BILINEAR):
        self.size = size
        self.interpolation = interpolation

    def __call__(self, img):
        """
        Args:
            img (PIL Image): Image to be scaled.

        Returns:
            PIL Image: Rescaled image.
        """
        width, height = img.size
        h, w = self.size

        v_scale = height / h
        h_scale = width / w
        scale = min(h_scale, v_scale)
        min_size = [scale * h, scale * w]

        img = F.center_crop(img, min_size)

        return F.resize(img, self.size, self.interpolation)

    def __repr__(self):
        interpolate_str = _pil_interpolation_to_str[self.interpolation]
        return self.__class__.__name__ + '(size={0}, interpolation={1})'.format(self.size, interpolate_str)




import numpy as np
import hyperchamber as hc
import inspect
import nashpy as nash
import hypergan as hg
import hyperchamber as hc
import sys
import gc
import os
import random

from hypergan.trainers.base_trainer import BaseTrainer

TINY = 1e-12

class GangTrainer(BaseTrainer):
    def create(self):
        config = self.config
        gan = self.gan
        loss = self.gan.loss
        d_vars = gan.d_vars()
        g_vars = gan.g_vars()
        self.reinit = tf.variables_initializer(d_vars+g_vars)
        self.priority_ds = []
        self.priority_gs = []


        self._delegate = self.gan.create_component(config.rbbr, d_vars=d_vars, g_vars=g_vars)
        self.optimizer = self._delegate

        print("VARS  ___", self.optimizer.variables())
        for var in self.optimizer.variables():
            if "generator" in var.name:
                g_vars += [var]
            elif "discriminator" in var.name:
                d_vars += [var]
            else:
                print("[WARNING] Unknown gang variable, not added to g_vars or d_vars: ", var)
        d_loss, g_loss = loss.sample
        if self.config.fitness_method == "wasserstein":
            self.gang_loss = -(loss.d_real-loss.d_fake)
        elif self.config.fitness_method == "least_squares":
            b = gan.loss.config.labels[1]
            a = gan.loss.config.labels[0]
            self.gang_loss = tf.sign(loss.d_fake + loss.d_real) * tf.square(loss.d_fake+loss.d_real)
        elif self.config.fitness_method == "least_squares2":
            self.gang_loss = tf.square(loss.d_fake)-tf.square(loss.d_real)
        elif self.config.fitness_method == "least_squares3":
            c = gan.loss.config.labels[2]
            b = gan.loss.config.labels[1]
            a = gan.loss.config.labels[0]
            self.gang_loss = tf.square(loss.d_fake)*tf.sign(loss.d_fake) + tf.square(loss.d_real)*tf.sign(d_real)
        elif self.config.fitness_method == "least_squares4":
            c = gan.loss.config.labels[2]
            b = gan.loss.config.labels[1]
            a = gan.loss.config.labels[0]
            self.gang_loss = tf.square((-loss.d_fake+1)/2) - tf.square((-loss.d_real+1)/2)

        elif self.config.fitness_method == 'double':
            self.gang_loss = [loss.d_fake+loss.d_real, loss.d_real-loss.d_fake]
        elif self.config.fitness_method == 'double_ragan':
            self.gang_loss = [0.7*(loss.d_fake-loss.d_real)+0.3*(loss.d_real-loss.d_fake), 
                    0.7*(loss.d_real-loss.d_fake)-0.3*(loss.d_fake-loss.d_real)]
        elif self.config.fitness_method == 'zero_raganls':
            c = gan.loss.config.labels[2]
            b = gan.loss.config.labels[1]
            a = gan.loss.config.labels[0]
            self.gang_loss = tf.square(loss.d_real-loss.d_fake-b) - tf.square(loss.d_fake - loss.d_real - b)
        elif self.config.fitness_method == 'zero_raganls2':
            c = gan.loss.config.labels[2]
            b = gan.loss.config.labels[1]
            a = gan.loss.config.labels[0]
            self.gang_loss = tf.square(loss.d_real-loss.d_fake-b)
        elif self.config.fitness_method == '-gl-dl':
            self.gang_loss = [-g_loss, -d_loss]
        elif self.config.fitness_method == 'gl-dl2':
            self.gang_loss = [-g_loss, g_loss]
        elif self.config.fitness_method == 'gldl':
            # breaks
            self.gang_loss = [g_loss, d_loss]
        elif self.config.fitness_method == 'double2':
            a = gan.loss.config.labels[0]
            b = gan.loss.config.labels[1]
            self.gang_loss = [tf.square(loss.d_fake-a)+tf.square(loss.d_real-a), tf.square(los++d_real-a)+tf.square(loss.d_fake-b)]
        elif self.config.fitness_method == "least_squares5":
            c = gan.loss.config.labels[2]
            b = gan.loss.config.labels[1]
            a = gan.loss.config.labels[0]
            self.gang_loss = tf.nn.relu(loss.d_real)*loss.d_fake
        elif self.config.fitness_method == "least_squares7":
            c = gan.loss.config.labels[2]
            b = gan.loss.config.labels[1]
            a = gan.loss.config.labels[0]
            self.gang_loss = (tf.nn.relu(loss.d_real)*tf.nn.relu(loss.d_fake))
        elif self.config.fitness_method == "least_squares8":
            c = gan.loss.config.labels[2]
            b = gan.loss.config.labels[1]
            a = gan.loss.config.labels[0]
            self.gang_loss = (loss.d_real-a)*(loss.d_fake-a)
        elif self.config.fitness_method == "least_squares9":
            c = gan.loss.config.labels[2]
            b = gan.loss.config.labels[1]
            a = gan.loss.config.labels[0]
            self.gang_loss = (loss.d_real-a)*(loss.d_fake-a)-(loss.d_real-b)*(loss.d_fake-b)
        elif self.config.fitness_method == "least_squares6":
            c = gan.loss.config.labels[2]
            b = gan.loss.config.labels[1]
            a = gan.loss.config.labels[0]
            self.gang_loss = tf.square(tf.nn.relu(loss.d_real))*tf.sign(loss.d_fake)*tf.square(loss.d_fake)

        elif self.config.fitness_method == "raleast_squares":
            c = gan.loss.config.labels[2]
            b = gan.loss.config.labels[1]
            a = gan.loss.config.labels[0]
            self.gang_loss = tf.square((-(loss.d_fake - loss.d_real)+1)/2) - tf.square((-(loss.d_real-loss.d_fake)+1)/2)
        elif self.config.fitness_method == "jsd":
            def mean_entropy(p, q):
                return p * (tf.log(p / (0.5*(p+q))))
            p = loss.d_fake
            q = loss.d_real
            self.gang_loss = -((mean_entropy(q, p) + mean_entropy(p, q)) / 2)

        elif self.config.fitness_method == "g_loss":
            self.gang_loss = g_loss
        elif self.config.fitness_method == "d_loss":
            self.gang_loss = d_loss
        elif self.config.fitness_method == "-g_loss":
            self.gang_loss = -g_loss
        elif self.config.fitness_method == 'ralsgan':
            b = gan.loss.config.labels[1]
            a = gan.loss.config.labels[0]
            self.gang_loss = 0.5*tf.square(loss.d_real - loss.d_fake - b) + 0.5*tf.square(loss.d_fake - loss.d_real - b)
        elif self.config.fitness_method == 'ragan':
            self.gang_loss = -tf.nn.sigmoid(loss.d_real-loss.d_fake)+ \
                         tf.nn.sigmoid(loss.d_fake-loss.d_real)
        elif self.config.fitness_method == 'ragan2':
            self.gang_loss = tf.log(tf.nn.sigmoid(loss.d_real)+TINY) - \
                         tf.log(tf.nn.sigmoid(loss.d_fake)+TINY)
        elif self.config.fitness_method == 'f-r':
            #self.gang_loss = tf.nn.sigmoid(loss.d_fake - loss.d_real)
            self.gang_loss = 2*tf.nn.sigmoid(2*(loss.d_fake-loss.d_real))-1
        elif self.config.fitness_method == 'f-r2':
            #self.gang_loss = tf.nn.sigmoid(loss.d_fake - loss.d_real)
            self.gang_loss = tf.nn.sigmoid((loss.d_fake))-tf.nn.sigmoid((loss.d_real))
        elif self.config.fitness_method == 'f-r3':
            #self.gang_loss = tf.nn.sigmoid(loss.d_fake - loss.d_real)
            self.gang_loss = tf.nn.sigmoid(loss.d_fake-loss.d_real)-tf.nn.sigmoid(loss.d_real-loss.d_fake)
        else:
            self.gang_loss = loss.d_fake - loss.d_real



        g_vars = list(g_vars)
        d_vars = list(d_vars)
        self.all_g_vars = g_vars
        self.all_d_vars = d_vars

        def random_like(x):
            shape = self.ops.shape(x)
            return tf.random_uniform(shape, minval=-0.01, maxval=0.01)

        self.ug = None#gan.session.run(g_vars)
        self.ud = None#gan.session.run(d_vars)
        self.pg = [tf.zeros_like(v) for v in g_vars]
        self._assign_g = [v.assign(pv) for v,pv in zip(g_vars, self.pg)]
        self.pd = [tf.zeros_like(v) for v in d_vars]
        self._assign_d = [v.assign(pv) for v,pv in zip(d_vars, self.pd)]
        self.pm = tf.zeros([1])
        self.assign_add_d = [v.assign(pv*self.pm+v) for v,pv in zip(d_vars, self.pd)]
        self.assign_add_g = [v.assign(pv*self.pm+v) for v,pv in zip(g_vars, self.pg)]
        self.mutate_g = [v.assign(random_like(v)+v) for v in d_vars]
        self.mutate_d = [v.assign(random_like(v)+v) for v in g_vars]

        self.sgs = []
        self.sds = []

        self.last_fitness_step = 0

    def required(self):
        return ""


    def rank_gs(self, gs):
        # todo fitness?
        return list(np.flip(gs, axis=0)) # most recent

    def rank_ds(self, ds):
        # todo fitness?
        return list(np.flip(ds, axis=0)) # most recent

    def destructive_mixture_g(self, priority_g):
        g_vars = self.all_g_vars
        self.gan.session.run(self._assign_g, {}) # zero
        for i, s in enumerate(self.sgs):
            self.add_g(priority_g[i], s)
        return self.gan.session.run(g_vars)

    def destructive_mixture_d(self, priority_d):
        d_vars = self.all_d_vars
        self.gan.session.run(self._assign_d, {}) # zero
        for i, s in enumerate(self.sds):
            self.add_d(priority_d[i], s)
        return self.gan.session.run(d_vars)

    def nash_memory(self, sg, sd, ug, ud):
        is_sd_nan = np.isnan(np.sum(np.sum(v) for v in sg))
        is_sg_nan = np.isnan(np.sum(np.sum(v) for v in sd))

        if is_sd_nan or is_sg_nan:
            print("NAN detected, falling back to best candidate")
            return [self.sgs[0], self.sds[0]]
        
        #zs = [ self.gan.session.run(self.gan.fitness_inputs()) for i in range(self.config.fitness_test_points or 10)]
        #xs = [ self.gan.session.run(self.gan.inputs.inputs()) for i in range(self.config.fitness_test_points or 10)]
        #self.xs = xs
        #self.zs = zs
        xs = []
        zs = []
        self.sgs = [sg] + self.sgs
        self.sds = [sd] + self.sds

        if isinstance(self.gang_loss, list):
            a = self.payoff_matrix(self.sgs, self.sds, xs, zs, self.gang_loss[0])
            b = self.payoff_matrix(self.sgs, self.sds, xs, zs, self.gang_loss[1])
            print("Payoffa:", a)
            print("Payoffb:", b)
        else:
            print("Calculating nash")
            a = self.payoff_matrix(self.sgs, self.sds, xs, zs)
            if np.min(a) == np.max(a) or np.isnan(np.sum(a)):
                print("WARNING: Degenerate game, skipping")
                print(a)
                self.priority_ds = list(np.zeros(len(self.sds)))
                self.priority_gs = list(np.zeros(len(self.sgs)))
                return [ug, ud]
            print("Payoff:", a)
            b = -a

        is_a_nan = np.isnan(np.sum(np.sum(v) for v in a))
        is_b_nan = np.isnan(np.sum(np.sum(v) for v in b))

        if is_a_nan or is_b_nan:
            print("NAN detected in payoff matrix, falling back to best candidate")
            self.sgs = self.sgs[1:]
            self.sds = self.sds[1:]
            return [self.sgs[0], self.sds[0]]
 
        if self.config.use_nash:
            priority_g, new_ug, priority_d, new_ud = self.nash_mixture_from_payoff(a, b, self.sgs, self.sds)
        elif self.config.use_crossover:
            priority_g = self.mixture_from_payoff(a, 1, self.sgs)
            priority_d = self.mixture_from_payoff(b, 0, self.sds)
        else:
            priority_g = self.mixture_from_payoff(a, 1, self.sgs)
            new_ug = self.destructive_mixture_g(priority_g)

            priority_d = self.mixture_from_payoff(b, 0, self.sds)
            new_ud = self.destructive_mixture_d(priority_d)

        memory_size = self.config.nash_memory_size or 10
        sorted_sgs = [[p, v] for p,v in zip(priority_g, self.sgs)]
        sorted_sds = [[p, v] for p,v in zip(priority_d, self.sds)]
        sorted_sgs.sort(key=lambda x: -x[0])
        sorted_sds.sort(key=lambda x: -x[0])
        print('mixture g:', [x[0] for x in sorted_sgs])
        print('mixture d:', [x[0] for x in sorted_sds])
        self.priority_gs = [x[0] for x in sorted_sgs]
        self.priority_ds = [x[0] for x in sorted_sds]
        sorted_sds = [s[1] for s in sorted_sds]
        sorted_sgs = [s[1] for s in sorted_sgs]
        self.sgs = sorted_sgs[:memory_size]
        self.sds = sorted_sds[:memory_size]
        self.priority_gs = self.priority_gs[:memory_size]
        self.priority_ds = self.priority_ds[:memory_size]

        if self.config.use_crossover:
            new_ug = self.crossover(self.sgs[0],self.sgs[1])
            new_ud = self.crossover(self.sds[0],self.sds[1])


        return [new_ug, new_ud]

    def crossover(self, s1, s2):
        strat = list(s1)
        for i,layer in enumerate(s2):
            mask = np.random.rand(*layer.shape)
            if self.config.crossover_random == None:
                mask = np.around(mask)
            strat[i]=mask*layer + (1-mask)*strat[i]
        return strat

    def softmax(self, x):
        e_x = np.exp(x - np.max(x))
        return e_x / e_x.sum(axis=0)
    def sumdiv(self, x):
        e_x = x
        return e_x / e_x.sum(axis=0)

    def nash_mixture_from_payoff(self, payoffa, payoffb, sgs, sds):
        config = self.config
        def _update_g(p):
            p = np.reshape(p, [-1])
            result = self.destructive_mixture_g(p)
            return p, result

        def _update_d(p):
            p = np.reshape(p, [-1])
            result = self.destructive_mixture_d(p)
            return p, result

        if self.config.nash_method == 'support':
            try:
                u = next(nash.Game(payoffa, payoffb).support_enumeration())
            except(StopIteration):
                print("Nashpy 'support' iteration failed.  Using 1,0,0...")
                u = [list(np.zeros(len(self.sds))), list(np.zeros(len(self.sgs)))]
                u[0][0]=1.
                u[1][0]=1.

        elif self.config.nash_method == 'lemke':
            u = next(nash.Game(payoffa, payoffb).lemke_howson_enumeration())

        else:
            try:
                u = next(nash.Game(payoffa, payoffb).vertex_enumeration())
            except(StopIteration, scipy.spatial.qhull.QhullError):
                print("Nashpy 'vertex' iteration failed.  Using 1,0,0...")
                u = [list(np.zeros(len(self.sds))), list(np.zeros(len(self.sgs)))]
                u[0][0]=1.
                u[1][0]=1.

        if len(u[0]) != len(self.sgs):
            return [None,None,None,None]
        p1, p1result = _update_g(u[0])
        p2, p2result = _update_d(u[1])

        return p1, p1result, p2, p2result


    def mixture_from_payoff(self, payoff, sum_dim, memory):
        u = np.sum(payoff, axis=sum_dim)
        u = self.softmax(u)
        u = np.reshape(u, [len(memory)])
        return u

    def payoff_matrix(self, sgs, sds, xs, zs, method=None):
        self._payoff_matrix = np.zeros([len(sgs), len(sds)])
        result = self._payoff_matrix
        for i, sg in enumerate(sgs):
            for j, sd in enumerate(sds):
                result[i, j]=self.fitness_score(sg, sd, xs, zs, method) # todo fitness ?
        return result

    def fitness_score(self, g, d, xs, zs, method=None):
        self.assign_gd(g,d)
        sum_fitness = 0
        test_points = self.config.fitness_test_points or 10
        if method == None:
            method = self.gang_loss
        for i in range(test_points):
            df, dr = self.gan.session.run([self.gan.loss.d_fake, self.gan.loss.d_real])
            fitness = self.gan.session.run(method)
            sum_fitness += np.average(fitness)
        #for x, z in zip(xs, zs):
        #    loss = self.loss or self.gan.loss
        #    feed_dict = {}
        #    for v, t in zip(x, self.gan.inputs.inputs()):
        #        feed_dict[t]=v
        #    for v, t in zip(z, self.gan.fitness_inputs()):
        #        feed_dict[t]=v
        #    fitness = self.gan.session.run([method], feed_dict)
        #    sum_fitness += np.average(fitness)

        sum_fitness /= float(test_points)
        return sum_fitness

    def assign_gd(self, g, d):
        self.assign_g(g)
        self.assign_d(d)

    def assign_g(self, g):
        fg = {}
        for v, t in zip(g, self.pg):
            fg[t] = v
        self.gan.session.run(self._assign_g, fg)

    def assign_d(self, d):
        fd = {}
        for v, t in zip(d, self.pd):
            fd[t] = v
        self.gan.session.run(self._assign_d, fd)

    def add_g(self, pm, g):
        fg = {}
        for v, t in zip(g, self.pg):
            fg[t] = v
        fg[self.pm] = np.reshape(pm, [1])
        self.gan.session.run(self.assign_add_g, fg)

    def add_d(self, pm, d):
        fd = {}
        for v, t in zip(d, self.pd):
            fd[t] = v
        fd[self.pm] = np.reshape(pm, [1])

        self.gan.session.run(self.assign_add_d, fd)

    def train_g_on_sds(self):
        gan = self.gan
        cd = gan.session.run(self._delegate.d_vars)
        gl = np.zeros(self._delegate.g_loss.shape)
        dl = np.zeros(self._delegate.d_loss.shape)
        for i,sd in enumerate(self.sds):
            p= self.priority_ds[i]
            if(p == 0):
                next
            self.assign_d(sd)

            _gl, _dl, *zs = gan.session.run([self._delegate.g_loss, self._delegate.d_loss]+gan.fitness_inputs())
            print("Train strategy", i, "P", p, "GL", _gl, "DL", _dl)
            gl += _gl * p
            dl += _dl * p
        feed_dict = {}
        for v, t in ([[gl*p, self._delegate.g_loss],[dl*p, self._delegate.d_loss]] + [ [v, t] for v, t in zip(zs, gan.fitness_inputs())]):
            feed_dict[t]=v
        _ = gan.session.run([self._delegate.g_optimizer], feed_dict)
        self.assign_d(cd)

    def train_d_on_sgs(self):
        gan = self.gan
        cg = gan.session.run(self._delegate.g_vars)
        for i,sg in enumerate(self.sgs):
            p= self.priority_gs[i]
            if(p == 0):
                next
            self.assign_g(sg)

            _, _gl, _dl, *zs = gan.session.run([self._delegate.d_optimizer, self._delegate.g_loss, self._delegate.d_loss]+gan.fitness_inputs())
            print("Train strategy", i, "P", p, "GL", _gl, "DL", _dl)

        self.assign_g(cg)

  
    def _step(self, feed_dict):
        gan = self.gan
        sess = gan.session
        config = self.config
        loss = gan.loss
        metrics = loss.metrics
        d_vars = self.all_d_vars
        g_vars = self.all_g_vars
        
        if self.ug == None:
            self.ug = gan.session.run(g_vars)
            self.ud = gan.session.run(d_vars)
            self.sgs.append(self.ug)
            self.sds.append(self.ud)
            self.priority_gs = [1]
            self.priority_ds = [1]
            if self.config.preload:
                for preload in self.config.preload:
                    print("Preloading ", preload)
                    default_save_path = os.path.abspath("saves/"+preload)
                    save_file = default_save_path + "/model.ckpt"
                    gan.session.run(tf.global_variables_initializer())
                    if not self.gan.load(save_file):
                        sys.exit("Could not load " + save_file)
                    print("Assigning to sds")
                    preload_ug = gan.session.run(g_vars)
                    preload_ud = gan.session.run(d_vars)
                    self.sgs.append(preload_ug)
                    self.sds.append(preload_ud)
                    self.priority_gs += [0]
                    self.priority_ds += [0]
 
                self.assign_gd(self.ug, self.ud)

        self._delegate.step(feed_dict)

        if config.train_g_on_sds and ((self._delegate.current_step+1) % (config.sds_steps or 100) == 0) and np.max(self.priority_ds) != 0:
            self.train_g_on_sds()
        if config.train_d_on_sgs and ((self._delegate.current_step+1) % (config.sgs_steps or 100) == 0) and np.max(self.priority_gs) != 0:
            self.train_d_on_sgs()
        
        #if self.last_fitness_step == self._delegate.current_step:
        #    return
        self.last_fitness_step=self._delegate.current_step
        #print("Step", self._delegate.current_step+1)
        if (gan.step_count+1) % (config.mix_steps or 100) == 0:
            sg = gan.session.run(g_vars)
            sd = gan.session.run(d_vars)
            if config.nash_memory:
                print("ENABLING NASH MEM", len(self.sgs))
                ug, ud = self.nash_memory(sg, sd, self.ug, self.ud)
                print("/ENABLING NASH MEM", len(self.sgs))
            else:
                decay = config.decay or 0.5
                ug = [ (o*decay + n*(1-decay)) for o, n in zip(sg, self.ug) ]
                ud = [ (o*decay + n*(1-decay)) for o, n in zip(sd, self.ud) ]

            if config.recreate:
                gan.train_coordinator.request_stop()
                gan.train_coordinator.join(gan.input_threads)
                gan.session.close()
                tf.reset_default_graph()
                #x = tf.contrib.graph_editor.copy(self.gan.inputs.x)
                #self.gan.inputs.x = x
                inputs = hg.inputs.image_loader.ImageLoader(gan.args.batch_size)
                inputs.create(gan.args.directory,
                      channels=gan.x_channels, 
                      format=gan.args.format,
                      crop=gan.args.crop,
                      width=gan.x_width,
                      height=gan.x_height,
                      resize=gan.args.resize)
                config_name = random.choice(self.config.mutations)

                newconfig_file = hg.Configuration.find(config_name+'.json')
                newconfig = hc.Selector().load(newconfig_file)

                newgan = self.gan.config['class'](config=newconfig, inputs=inputs)
                newgan.args = gan.args
                newgan.x_width = gan.x_width
                newgan.x_height = gan.x_height
                newgan.x_channels = gan.x_channels
                newgan.cli = self.gan.cli
                newgan.cli.sampler = None
                gan.cli.sampler = None
                newgan.trainer.sds = self.sds
                newgan.trainer.sgs = self.sgs
                newgan.train_coordinator = tf.train.Coordinator()
                self.sds = None
                self.sgs = None
                self.ug = None
                self.ud = None
                for c in gan.components:
                    c.gan = None
                
                def out(o):
                    return str(type(o))
                self.gan=None
                gc.collect()
                #for v in gc.get_referrers(gan):
                #    try:
                #        print(v.f_locals['self'].__class__) 
                #    except AttributeError:
                #        print("!", v) 
                #print(sys.getrefcount(gan), len(gc.get_referrers(gan)))

                gan.destroy=True
                gan.newgan=newgan
                gan=None
                gc.collect()
                newgan.input_threads = tf.train.start_queue_runners(sess=newgan.session, coord=newgan.train_coordinator)
                newgan.trainer.assign_gd(ug, ud)
                return

            self.assign_gd(ug, ud)

            if self.config.mutate_child and np.random.rand() > (self.config.mutation_chance or 0.7):
                print("Mutating child")
                self.gan.session.run([self.mutate_d, self.mutate_g])

            self.ug = gan.session.run(self.all_g_vars)
            self.ud = gan.session.run(self.all_d_vars)
            if self.current_step < (config.reset_before_step or 0):
                gan.session.run(tf.global_variables_initializer())










# Symplectic Gradient Adjustment
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from tensorflow.python.ops import control_flow_ops
from tensorflow.python.ops import math_ops
from tensorflow.python.ops import state_ops
from tensorflow.python.framework import ops
from tensorflow.python.training import optimizer
import hyperchamber as hc
import inspect

class SgaOptimizer(optimizer.Optimizer):
  def __init__(self, learning_rate=0.001, p=0.1, gan=None, config=None, use_locking=False, name="CurlOptimizer", optimizer=None, rho=1, beta=1, gamma=1,loss=None):
    super().__init__(use_locking, name)
    self._beta = beta
    self._rho = rho
    self._gamma = gamma
    self.gan = gan
    self.config = config
    self._lr_t = learning_rate
    self.loss = loss
    optimizer["loss"] = loss

    self.optimizer = self.gan.create_optimizer(optimizer)
 
  def _prepare(self):
    super()._prepare()
    self.optimizer._prepare()

  def _create_slots(self, var_list):
    super()._create_slots(var_list)
    self.optimizer._create_slots(var_list)

  def _apply_dense(self, grad, var):
    return self.optimizer._apply_dense(grad, var)


  def fwd_gradients(self, ys, xs, grad_xs=None, stop_gradients=None, colocate_gradients_with_ops=True):
    us = [tf.zeros_like(y) + float('nan') for y in ys]
    dydxs = tf.gradients(ys, xs, grad_ys=us,stop_gradients=stop_gradients,colocate_gradients_with_ops=colocate_gradients_with_ops)
    dydxs = [tf.zeros_like(x) if dydx is None else dydx for x,dydx in zip(xs,dydxs)]
    dysdx = tf.gradients(dydxs, us, grad_ys=grad_xs, colocate_gradients_with_ops=colocate_gradients_with_ops)
    return dysdx
  def apply_gradients(self, grads_and_vars, global_step=None, name=None):
    ws = [v for _,v in grads_and_vars]
    grads = [g for g,_ in grads_and_vars]
    self._prepare()

    jac_vec = self.fwd_gradients(grads,ws, grad_xs=grads,stop_gradients=ws)
    jac_vec = [tf.zeros_like(x) if dydx is None else dydx for x,dydx in zip(ws,jac_vec)]
    jac_tran_vec = tf.gradients(grads, ws, grad_ys=grads, stop_gradients=ws)
    jac_tran_vec = [tf.zeros_like(x) if dydx is None else dydx for x,dydx in zip(ws,jac_tran_vec)]
    at_xi = [(ht-h)*0.5 for (h,ht) in zip(jac_vec, jac_tran_vec)]


    if self.config.minus:
        new_grads = [g-a for g,a in zip(grads, at_xi)]
    else:
        new_grads = [g+a for g,a in zip(grads, at_xi)]
    grads_and_vars2 = zip(new_grads, ws)
    op8 = self.optimizer.apply_gradients(list(grads_and_vars2).copy(), global_step=global_step, name=name)
    with tf.get_default_graph().control_dependencies([op8]):
        return tf.no_op()


  def _apply_sparse(self, grad, var):
    raise NotImplementedError("Sparse gradient updates are not supported.")
  def variables(self):
      return super().variables() + self.optimizer.variables()



#From https://gist.github.com/EndingCredits/b5f35e84df10d46cfa716178d9c862a3
import math
import torch
from torch.optim import Optimizer


class Curl(Optimizer):
    def __init__(self, params, lr=1e-3, optimizer=None):
        if not 0.0 <= lr:
            raise ValueError("Invalid learning rate: {}".format(lr))
        if optimizer is None:
            raise ValueError("Requires optimizer")
        defaults = dict(lr=lr)

        super(Curl, self).__init__(params, defaults)

    def __setstate__(self, state):
        super(Adam, self).__setstate__(state)
        for group in self.param_groups:
            group.setdefault('amsgrad', False)

    def step(self, closure=None):
#class CurlOptimizer(optimizer.Optimizer):
#  def __init__(self, learning_rate=0.00001, p=0.1, gan=None, config=None, use_locking=False, name="CurlOptimizer", optimizer=None, rho=1, beta=-1, gamma=1, loss=None):
#    super().__init__(use_locking, name)
#    self._beta = gan.configurable_param(beta)
#    self._rho = rho
#    self._gamma = gamma
#    self.gan = gan
#    self.config = config
#    self._lr_t = learning_rate
#    self.g_rho = gan.configurable_param(self.config.g_rho or 1)
#    self.d_rho = gan.configurable_param(self.config.d_rho or 1)
#
#    optimizer['loss'] = loss
#    self.optimizer = self.gan.create_optimizer(optimizer)
# 
#  def _prepare(self):
#    super()._prepare()
#    self.optimizer._prepare()
#
#  def _create_slots(self, var_list):
#    super()._create_slots(var_list)
#    self.optimizer._create_slots(var_list)
#
#  def _apply_dense(self, grad, var):
#    return self.optimizer._apply_dense(grad, var)
#
#  def apply_gradients(self, grads_and_vars, global_step=None, name=None):
#    d_vars = []
#    g_vars = []
#    d_grads = []
#    g_grads = []
#
#    for grad,var in grads_and_vars:
#        if var in self.gan.d_vars():
#            d_vars += [var]
#            d_grads += [grad]
#        elif var in self.gan.g_vars():
#            g_vars += [var]
#            g_grads += [grad]
#        else:
#            raise("Couldn't find var in g_vars or d_vars")
#
#    var_list = d_vars + g_vars
#
#    with ops.init_scope():
#        slots_list = []
#        if self.config.include_slots:
#            for name in self.optimizer.get_slot_names():
#                for var in self.optimizer.variables():
#                    slots_list.append(self._zeros_slot(var, "curl", "curl"))
#    self._prepare()
#
#    def _name(post, s):
#        ss = s.split(":")
#        return ss[0] + "_" + post + "_dontsave"
#
#
#    v1 = [tf.Variable(v, name=_name("curl",v.name)) for v in var_list]
#    self.gan.ops.weights += v1
#    slots_list = []
#    slots_vars = []
#    if self.config.include_slots:
#        for name in self.optimizer.get_slot_names():
#            for var in self.optimizer.variables():
#                slots_vars += [var]
#                slots_list.append(self._zeros_slot(var, "curl", "curl"))
#
#
#    restored_vars = var_list + slots_vars
#    tmp_vars = v1 + slots_list
#    # store variables for resetting
#
#    if self.config.beta_type == 'sga':
#        Jgrads = tf.gradients(d_grads, d_vars, grad_ys=d_grads, stop_gradients=d_vars) + [tf.zeros_like(g) for g in g_vars]
#    elif self.config.beta_type == 'magnitude':
#        consensus_reg = [tf.square(g) for g in d_grads if g is not None]
#        Jgrads = tf.gradients(consensus_reg, d_vars) + [tf.zeros_like(g) for g in g_vars]
#    else:
#        consensus_reg = 0.5 * sum(
#                tf.reduce_sum(tf.square(g)) for g in d_grads if g is not None
#        )
#        Jgrads = tf.gradients(consensus_reg, d_vars, stop_gradients=d_vars) + [tf.zeros_like(g) for g in g_vars]
#
#    g1s = d_grads + g_grads
#
#    op1 = tf.group(*[tf.assign(w, v) for w,v in zip(tmp_vars, restored_vars)]) # store variables
#
#    with tf.get_default_graph().control_dependencies([op1]):
#        # store g2
#        op3 = tf.group(*[tf.assign_sub(v, self._lr_t*grad) for grad,v in grads_and_vars])
#        with tf.get_default_graph().control_dependencies([op3]):
#
#            def curlcombine(g1,g2,_v1,_v2,curl,rho):
#                stepsize = self._lr_t
#                if curl == "mirror":
#                    return self._gamma*(g1 + 2*g2)
#                else:
#                    return self._gamma*g1-rho*(g2-g1)/stepsize
#            g2s = tf.gradients(self.gan.trainer.d_loss, d_vars) + tf.gradients(self.gan.trainer.g_loss, g_vars)
#            if self.config.form == 'central':
#                def central_step():
#                    # restore v1, slots
#                    op5 = tf.group(*[ tf.assign(w,v) for w,v in zip(restored_vars, tmp_vars)])
#                    with tf.get_default_graph().control_dependencies([op5]):
#                        back =  tf.group(*[tf.assign_sub(v, -self._lr_t*grad) for grad,v in grads_and_vars])
#                        with tf.get_default_graph().control_dependencies([back]):
#                            return tf.gradients(self.gan.trainer.d_loss, d_vars) + tf.gradients(self.gan.trainer.g_loss, g_vars)
#                def curlcombinecentral(g1,g2,_v1,_v2,curl,rho):
#                    #stepsize = (_v2-_v1)/(g1+1e-8)
#                    stepsize = self._lr_t
#                    if curl == "mirror":
#                        return self._gamma*(g1 + 2*g2)
#                    else:
#                        return self._gamma*g1-rho*(g2-g1)/(2*stepsize)
#
#                g1s  = central_step()
#                g3s = [curlcombinecentral(g1,g2,v1,v2,self.config.d_curl,self.d_rho) if v2 in d_vars else curlcombinecentral(g1,g2,v1,v2,self.config.g_curl,self.g_rho) for g1,g2,v1,v2 in zip(g1s,g2s,v1,var_list)]
#            else:
#                #forward
#                g3s = [curlcombine(g1,g2,v1,v2,self.config.d_curl,self.d_rho) if v2 in d_vars else curlcombine(g1,g2,v1,v2,self.config.g_curl,self.g_rho) for g1,g2,v1,v2 in zip(g1s,g2s,v1,var_list)]
#            # restore v1, slots
#            op5 = tf.group(*[ tf.assign(w,v) for w,v in zip(restored_vars, tmp_vars)])
#            with tf.get_default_graph().control_dependencies([op5]):
#                flin = []
#                for grad, jg in zip(g3s, Jgrads):
#                    if jg is None or self._beta <= 0:
#                        flin += [grad]
#                    else:
#                        flin += [grad + jg * self._beta]
#
#                if self.config.orthonormal:
#                    shapes = [self.gan.ops.shape(l) for l in flin]
#                    u = [tf.reshape(l, [-1]) for l in flin[:len(d_vars)]]
#                    v = [tf.reshape(l, [-1]) for l in Jgrads[:len(d_vars)]]
#                    
#                    def proj(u, v,shape):
#                        dot = tf.tensordot(v, u, 1) / (tf.square(u)+1e-8)
#                        dot = tf.maximum(-1.0, dot)
#                        dot = tf.minimum(1.0, dot)
#                        dot = dot * u
#                        dot = tf.reshape(dot, shape)
#                        return dot
#                    proj_u1_v2 = [proj(_u, _v, _s) for _u, _v, _s in zip(u, v, shapes)]
#                    flin = [_flin + self.gan.configurable_param(self.config.ortholambda) * proj for _flin, proj in zip(flin, proj_u1_v2)] + flin[len(d_vars):]
#
#                step3 = list(zip(flin, var_list))
#                op6 = self.optimizer.apply_gradients(step3.copy(), global_step=global_step, name=name)
#
#                with tf.get_default_graph().control_dependencies([op6]):
#                    return tf.no_op()
#
#  def _apply_sparse(self, grad, var):
#    raise NotImplementedError("Sparse gradient updates are not supported.")
#  def variables(self):
#      return self.optimizer.variables()



"""
Opens a window and communicates over it on ui events
"""
import numpy as np
import os
import contextlib
import time
import threading
import random
from queue import Queue
import queue


class ThreadedTkViewerUI:
    def __init__(self, title="HyperGAN", viewer_size=1, enabled=True):
        self.screen = None
        self.title = title
        self.viewer_size = viewer_size
        self.enable_menu = True

    def sample(self, gan, image):
        original_image = image
        if len(np.shape(image)) == 2:
            s = np.shape(image)
            image = np.reshape(image, [s[0], s[1], 1])
            image = np.tile(image, [1,1,3])
        image = np.transpose(image, [1, 0,2])

        self.size = [int(image.shape[0] * self.viewer_size), int(image.shape[1] * self.viewer_size)]

        def _refresh_sample(*args):
          gan.cli.should_sample=True

        if not self.screen:

            with contextlib.redirect_stdout(None):
                import pygame

            import tkinter as tk
            import tkinter.ttk
            class ResizableFrame(tk.Frame):
                def __init__(self,parent,tkviewer=None,**kwargs):
                    surface = kwargs.pop('surface')
                    tk.Frame.__init__(self,parent,**kwargs)
                    self.parent = parent
                    self.bind("<Configure>", self.on_resize)
                    self.height = kwargs['height']
                    self.width = kwargs['width']
                    self.surface = surface
                    self.tkviewer = tkviewer
                    self.aspect_ratio = float(self.width)/float(self.height)

                def on_resize(self,event):
                    self._update(event.width, event.height)
                    self.enforce_aspect_ratio(event.width, event.height)

                def _update(self, width, height):
                    self.width = width
                    self.height = height
                    self.config(width=self.width, height=self.height)
                    self.tkviewer.size = [self.width, self.height]
                    self.tkviewer.screen = self.tkviewer.pg.display.set_mode(self.tkviewer.size,self.tkviewer.pg.RESIZABLE)
                    self.tkviewer.screen.fill((0, 0, 0))
                    self.surface = self.tkviewer.pg.Surface([image.shape[0],image.shape[1]])
                    self.tkviewer.pg.surfarray.blit_array(self.surface, image[:,:,:3])
                    self.tkviewer.screen.blit(self.tkviewer.pg.transform.scale(self.surface,self.tkviewer.size),(0,0))
                    self.tkviewer.pg.display.flip()

                def resize(self, size):
                    pass
                    #self.parent.geometry(str(size[0])+"x"+str(size[1]))
                    #self._update(size[0], size[1])

                def enforce_aspect_ratio(self, width, height):
                    desired_width = width
                    desired_height = int(width / self.aspect_ratio)

                    if desired_height > height:
                        desired_height = height
                        desired_width = int(height * self.aspect_ratio)

                    self.config(width=desired_width, height=desired_height)
                    self.tkviewer.size = [desired_width, desired_height]
                    self.tkviewer.screen = self.tkviewer.pg.display.set_mode(self.tkviewer.size,self.tkviewer.pg.RESIZABLE)


            self.pg = pygame
            self.tk = tk
            self.surface = self.pg.Surface([image.shape[0],image.shape[1]])
            root = tk.Tk(className=self.title)
            self.resizable_frame = ResizableFrame(root, width=self.size[0], height=self.size[1], tkviewer=self, surface=self.surface)
            root.rowconfigure(0,weight=1)
            root.rowconfigure(1,weight=1)
            root.columnconfigure(0,weight=1)
            root.columnconfigure(1,weight=1)
            self.resizable_frame.pack(expand=tk.YES, fill=tk.BOTH)

            def _save_model(*args):
                gan.save(gan.cli.save_file)

            def _exit(*args):
                gan.exit()

            def _select_sampler(gan, name, value, submenu):
                def _select_sampler_proc():
                    gan.cli.sampler = gan.sampler_for(name)(gan, samples_per_row=gan.cli.args.width)
                    gan.cli.sample(False)
                    _refresh_sampler_submenu(submenu)
                return _select_sampler_proc

            def _refresh_sampler_submenu(submenu):
                if submenu.count > 0:
                    submenu.delete(0, submenu.count)

                for (k, v) in sorted(gan.get_registered_samplers().items()):
                    showall = tk.BooleanVar()
                    showall.set(gan.selected_sampler == k)
                    if v.compatible_with(gan):
                        state = tk.NORMAL
                    else:
                        state = tk.DISABLED

                    submenu.add_checkbutton(label=k, onvalue=True, offvalue=False, variable=showall, command=_select_sampler(gan, k, showall, submenu), state=state)
                num_samplers = len(gan.get_registered_samplers())

                submenu.count = num_samplers


            def _create_status_bar(root):
                statusbar = tk.Frame(root, height=24)
                statusbar.pack(side=tk.BOTTOM, fill=tk.X)

                label_training = tk.Label(statusbar, text="Training", font=12)
                label_training.grid(row=0,column=0) 
                sep = tkinter.ttk.Separator(statusbar, orient=tk.VERTICAL).grid(column=1, row=0, sticky='ns')
                label = tk.Label(statusbar, text="Starting", font=12)
                label.grid(row=0, column=2) 
                def __update_step():
                    label['text']=("Step " + str(int(gan.steps[0])))
                    root.after(1000, __update_step)


                __update_step()
                return statusbar

            menubar = tk.Menu(root)
            filemenu = tk.Menu(menubar, tearoff=0)
            filemenu.add_command(label="Save", command=_save_model, underline=0, accelerator="Ctrl+s")

            filemenu.add_separator()

            samplemenu = tk.Menu(menubar, tearoff=0)
            samplemenu.add_command(label="Refresh", command=_refresh_sample, underline=0, accelerator="Ctrl+r")

            filemenu.add_command(label="Save and Exit", command=_exit, underline=10, accelerator="Ctrl+q")
            menubar.add_cascade(label="File", menu=filemenu, underline=0)
            menubar.add_cascade(label="Sample", menu=samplemenu, underline=0)
            samplermenu = tk.Menu(samplemenu)
            samplemenu.add_cascade(label="Sampler", menu=samplermenu, underline=0)
            samplermenu.count = 0
            _refresh_sampler_submenu(samplermenu)

            root.bind_all("<Control-q>", _exit)
            root.bind_all("<Control-r>", _refresh_sample)
            root.bind_all("<Control-s>", _save_model)


            if self.enable_menu:
                root.config(menu=menubar)
                self.status_bar = _create_status_bar(root)
            else:
                self.status_bar = None

            # Tell pygame's SDL window which window ID to use
            os.environ['SDL_WINDOWID'] = str(self.resizable_frame.winfo_id())
            # Show the window so it's assigned an ID.
            root.update()
            self.root = root

            # Usual pygame initialization
            if self.viewer_size <= 0:
                self.viewer_size = 0.1
            self.aspect_w = image.shape[1] / image.shape[0]
            self.aspect_h = image.shape[0] / image.shape[1]
            self.temp_size = self.size
            self.screen = self.pg.display.set_mode(self.size,self.pg.RESIZABLE)
            self.pg.display.set_caption(self.title)

            root.title(self.title)
            root.wm_title(self.title)
            self.resizable_frame.winfo_toplevel().title(self.title)
            self.clock = self.pg.time.Clock()

        padw = 0
        padh = 0
        if original_image.shape[0] > original_image.shape[1]:
            padh = (original_image.shape[0] - original_image.shape[1])//2

        if original_image.shape[1] > original_image.shape[0]:
            padw = (original_image.shape[1] - original_image.shape[0])//2
        pad_image = np.pad(original_image, [(padw, padw), (padh,padh), (0,0)], 'constant')
        w = pad_image.shape[0]
        h = pad_image.shape[1]
        #xdata = b'P6 ' + str(w).encode() + b' ' + str(h).encode() + b' 255 ' + pad_image.tobytes()
        #tk_image = self.tk.PhotoImage(data=xdata, format="PPM", width=w, height=h)
        #self.root.tk.call('wm', 'iconphoto', self.root._w, tk_image.subsample(max(1, w//256), max(1, h//256)))

        if not hasattr(self, 'surface') or image.shape[0] != self.surface.get_width() or image.shape[1] != self.surface.get_height():
            self.surface = self.pg.Surface([image.shape[0],image.shape[1]])

        self.screen.fill((0, 0, 0))
        self.pg.surfarray.blit_array(self.surface, image[:,:,:3])
        self.screen.blit(self.pg.transform.scale(self.surface,self.size),(0,0))
        self.clock.tick_busy_loop(60)
        self.pg.display.flip()


    def event(self, gan, name, data):
        self.get_queue().put({"gan": gan, "name": name, "data": data})

    def get_queue(self):
        if hasattr(self, 'queue'):
            return self.queue
        self.queue = Queue()
        return self.queue

    def process(self, thread):
        while not thread.stopped():
            try:
                msg = self.get_queue().get(0)
                name = msg["name"]
                data = msg["data"]
                gan = msg["gan"]
                if name == "sample":
                    self.sample(gan, data)
                else:
                    print("Unknown event", msg)
            except queue.Empty:
                time.sleep(1.0/60.0)
                if hasattr(self, 'root'):
                    self.root.update()

class StoppableThread(threading.Thread):
    """Thread class with a stop() method. The thread itself has to check
    regularly for the stopped() condition."""

    def __init__(self, *args, **kwargs):
        self.ui = args[0]
        args = args[1:]
        super(StoppableThread, self).__init__(*args, **kwargs)
        self._stop_event = threading.Event()

    def stop(self):
        self._stop_event.set()

    def run(self):
        return self.ui.process(self)

    def stopped(self):
        return self._stop_event.is_set()

class ThreadedTkViewer:
    def __init__(self, title="HyperGAN", viewer_size=1, enabled=True):
        self.enabled = enabled
        self.ui = ThreadedTkViewerUI(title=title, viewer_size=viewer_size, enabled=enabled)
        self.ui_thread = StoppableThread(self.ui)
        self.ui_thread.start()

    def close(self):
        self.ui_thread.stop()

    def update(self, gan, image):
        if not self.enabled: return

        self.ui.event(gan, "sample", image)

    def set_options(self, enable_menu = None, title = None, viewer_size = None, enabled = None, zoom = None):
        self.enabled = enabled
        self.ui.enable_menu = enable_menu
        self.ui.title = title
        self.ui.viewer_size = viewer_size
        self.ui.zoom = zoom

# todo

# closing worker ui thread
# why does sampling repeatedly slow?
# changing samplers



import copy
import hyperchamber as hc
import inspect
import math
import operator
import os
import re
from functools import reduce

import pyparsing
import hypergan
import torch
import torch.nn as nn

from .gan_component import GANComponent
from hypergan.gan_component import ValidationException
from hypergan.layer_shape import LayerShape

from hypergan.modules.adaptive_instance_norm import AdaptiveInstanceNorm
from hypergan.modules.attention import Attention
from hypergan.modules.const import Const
from hypergan.modules.learned_noise import LearnedNoise
from hypergan.modules.modulated_conv2d import ModulatedConv2d, Blur, EqualLinear
from hypergan.modules.multi_head_attention import MultiHeadAttention
from hypergan.modules.reshape import Reshape
from hypergan.modules.no_op import NoOp
from hypergan.modules.scaled_conv2d import ScaledConv2d
from hypergan.modules.variational import Variational
from hypergan.modules.pixel_norm import PixelNorm

import torchvision
import hypergan as hg

class ConfigurableComponent(GANComponent):
    def __init__(self, gan, config, input=None, input_shape=None, context_shapes = {}):
        self.current_size = LayerShape(gan.channels(), gan.height(), gan.width())
        if isinstance(input, GANComponent):
            if hasattr(input, 'current_height'):
                self.current_size = LayerShape(input.current_channels, input.current_height, input.current_width)
            elif hasattr(input, 'current_channels'):
                self.current_size = LayerShape(input.current_channels)
            else:
                self.current_size = input.current_size
        if input_shape is not None:
            self.current_size = LayerShape(*input_shape)
        self.layers = []
        self.layer_shapes = []
        self.untrainable_parameters = set()
        self.layer_output_sizes = {}
        self.nn_layers = []
        self.layer_options = {}
        self.parsed_layers = []
        self.parser = hypergan.parser.Parser()
        self.context_shapes = context_shapes
        self.layer_ops = {**self.activations(),
            "add": hg.layers.Add,
            "cat": hg.layers.Cat,
            "channel_attention": hg.layers.ChannelAttention,
            "ez_norm": hg.layers.EzNorm,
            "mul": hg.layers.Mul,
            "multi_head_attention2": hg.layers.MultiHeadAttention, #TODO rename
            "pixel_shuffle": hg.layers.PixelShuffle,
            "residual": hg.layers.Residual,
            "resizable_stack": hg.layers.ResizableStack,
            "segment_softmax": hg.layers.SegmentSoftmax,
            "upsample": hg.layers.Upsample,

            #easy to convert
            "dropout": self.layer_dropout,
            "identity": self.layer_identity,
            "flatten": self.layer_flatten,
            "layer": self.layer_layer,
            "pretrained": self.layer_pretrained,
            "avg_pool": self.layer_avg_pool,#TODO handle dims
            "pad": self.layer_pad,
            "reshape": self.layer_reshape,
            "split": self.layer_split,

            #hard to convert
            "adaptive_avg_pool": self.layer_adaptive_avg_pool,
            "adaptive_avg_pool1d": self.layer_adaptive_avg_pool1d,
            "adaptive_avg_pool3d": self.layer_adaptive_avg_pool3d,
            "adaptive_instance_norm": self.layer_adaptive_instance_norm,
            "attention": self.layer_attention,
            "batch_norm": self.layer_batch_norm,
            "batch_norm1d": self.layer_batch_norm1d,
            "blur": self.layer_blur,
            "const": self.layer_const,
            "conv": self.layer_conv,
            "conv1d": self.layer_conv1d,
            "conv2d": self.layer_conv2d,
            "conv3d": self.layer_conv3d,
            "deconv": self.layer_deconv,
            "equal_linear": self.layer_equal_linear,
            "instance_norm": self.layer_instance_norm,
            "instance_norm1d": self.layer_instance_norm1d,
            "instance_norm3d": self.layer_instance_norm3d,
            "latent": self.layer_latent,
            "layer_norm": self.layer_norm,
            "learned_noise": self.layer_learned_noise,
            "linear": self.layer_linear,
            "modulated_conv2d": self.layer_modulated_conv2d,
            "module": self.layer_module,
            "multi_head_attention": self.layer_multi_head_attention,
            "pixel_norm": self.layer_pixel_norm,
            "resize_conv": self.layer_resize_conv,
            "resize_conv2d": self.layer_resize_conv2d,
            "resize_conv1d": self.layer_resize_conv1d,
            "scaled_conv2d": self.layer_scaled_conv2d,
            "subpixel": self.layer_subpixel,
            "vae": self.layer_vae
            #"linear_attention": hg.layers.LinearAttention,
            #"make2d": self.layer_make2d,
            #"make3d": self.layer_make3d,
            # "crop": self.layer_crop,
            # "dropout": self.layer_dropout,
            # "noise": self.layer_noise, #TODO
            #"attention": self.layer_attention, #TODO
            #"const": self.layer_const, #TODO
            #"gram_matrix": self.layer_gram_matrix, #TODO
            #"image_statistics": self.layer_image_statistics, #TODO
            #"knowledge_base": self.layer_knowledge_base, #TODO
            #"layer_norm": self.layer_layer_norm,#TODO
            #"mask": self.layer_mask,#TODO
            #"match_support": self.layer_match_support,#TODO
            #"minibatch": self.layer_minibatch,#TODO
            #"pixel_norm": self.layer_pixel_norm,#TODO
            #"progressive_replace": self.layer_progressive_replace,#TODO
            #"reduce_sum": self.layer_reduce_sum,#TODO might want to just do "reduce sum" instead
            #"relational": self.layer_relational,#TODO
            #"unpool": self.layer_unpool, #TODO https://arxiv.org/abs/1505.04366
            #"squash": self.layer_squash, #TODO
            #"tensorflowcv": self.layer_tensorflowcv, #TODO layer torchvision instead?
            #"turing_test": self.layer_turing_test, #TODO
            #"two_sample_stack": self.layer_two_sample_stack, #TODO
            #"zeros": self.layer_zeros, #TODO
            #"zeros_like": self.layer_zeros_like #TODO
            }
        self.named_layers = {}
        if not hasattr(gan, "named_layers"):
            gan.named_layers = {}
        self.subnets = hc.Config(hc.Config(config).subnets or {})
        GANComponent.__init__(self, gan, config)
        self.const_two = torch.Tensor([2.0]).float()[0].cuda()
        self.const_one = torch.Tensor([1.0]).float()[0].cuda()

    def required(self):
        return "layers".split()

    def layer(self, name):
        if name in self.gan.named_layers:
            return self.gan.named_layers[name] 
        if name in self.named_layers:
            return self.named_layers[name]
        return None

    def create(self):
        for layer in self.config.layers:
            net = self.create_parsed_layer(layer)
            self.nn_layers.append(net)

        self.net = nn.ModuleList(self.nn_layers)

    def create_parsed_layer(self, layer_defn):
        config = self.config
        parsed, layer = self.parse_layer(layer_defn)
        self.parsed_layers.append(parsed)
        self.layer_shapes.append(self.current_size)
        return layer

    def parse_layer(self, layer_defn):
        print("Parsing layer:", layer_defn)
        parsed = self.parser.parse_string(layer_defn)
        parsed.parsed_options = hc.Config(parsed.options)
        parsed.layer_defn = layer_defn
        print("Parsed layer:", parsed.to_list())
        layer = self.build_layer(parsed.layer_name, parsed.args, parsed.parsed_options)
        return parsed, layer

    def build_layer(self, op, args, options):
        if self.layer_ops[op]:
            try:
                is_hg_module = issubclass(self.layer_ops[op], hg.Layer)
            except TypeError:
                is_hg_module = False

            if is_hg_module:
                net = self.layer_ops[op](self, args, options)
                self.current_size = net.output_size()
            elif isinstance(self.layer_ops[op], nn.Module):
                net = self.layer_ops[op]
            else:
                #before_count = self.count_number_trainable_params()
                print("Size before: ", self.current_size.dims)
                net = self.layer_ops[op](None, args, options)
                print("Size after: ", self.current_size.dims)
            if 'name' in options:
                self.set_layer(options['name'], net)
            if options.trainable == False:
                self.untrainable_parameters = self.untrainable_parameters.union(set(net.parameters()))
            return net

            #after = self.variables()
            #new = set(after) - set(before)
            #for j in new:
            #    self.layer_options[j]=options
            #after_count = self.count_number_trainable_params()
            #if not self.ops._reuse:
            #    if net == None:
            #        print("[Error] Layer resulted in null return value: ", op, args, options)
            #        raise ValidationException("Configurable layer is null")
            #    print("layer: ", self.ops.shape(net), op, args, after_count-before_count, "params")
        else:
            print("ConfigurableComponent: Op not defined", op)

    def set_layer(self, name, net):
        self.gan.named_layers[name] = net
        self.named_layers[name]     = net
        self.layer_output_sizes[name] = self.current_size

    def activations(self):
        return {
            "celu": nn.CELU(),
            "gelu": nn.GELU(),
            "lrelu": nn.LeakyReLU(0.2),
            "prelu": nn.PReLU(),
            "relu": nn.ReLU(),
            "relu6": nn.ReLU6(),
            "selu": nn.SELU(),
            "sigmoid": nn.Sigmoid(),
            "softplus": nn.Softplus(),
            "softshrink": nn.Softshrink(),
            "softsign": nn.Softsign(),
            "hardtanh": nn.Hardtanh(),
            "tanh": nn.Tanh(),
            "tanhshrink": nn.Tanhshrink()
        }

    def layer_dropout(self, net, args, options):
        return nn.Dropout2d(float(args[0]))

    def layer_identity(self, net, args, options):
        return NoOp()

    def layer_equal_linear(self, net, args, options):
        lr_mul = 1
        if options.lr_mul is not None:
            lr_mul = options.lr_mul
        result = EqualLinear(self.current_size.size(), args[0], lr_mul=lr_mul)
        self.current_size = LayerShape(args[0])
        return result

    def get_same_padding(self, input_rows, filter_rows, stride, dilation):
        out_rows = (input_rows + stride - 1) // stride
        return max(0, (out_rows - 1) * stride + (filter_rows - 1) * dilation + 1 - input_rows) // 2

    def layer_const(self, net, args, options):
        return Const(*self.current_size.dims)

    #from https://discuss.pytorch.org/t/utility-function-for-calculating-the-shape-of-a-conv-output/11173/3
    def conv_output_shape(self, h_w, kernel_size=1, stride=1, pad=0, dilation=1):
        if type(kernel_size) is not tuple:
            kernel_size = (kernel_size, kernel_size)
        h = math.floor( ((h_w[0] + (2 * pad) - ( dilation * (kernel_size[0] - 1) ) - 1 )/ stride) + 1)
        w = math.floor( ((h_w[1] + (2 * pad) - ( dilation * (kernel_size[1] - 1) ) - 1 )/ stride) + 1)
        return h, w

    def layer_conv(self, net, args, options):
        return self.layer_conv2d(net, args, options)

    def layer_conv2d(self, net, args, options):
        if len(args) > 0:
            channels = args[0]
        else:
            channels = self.current_size.channels
        options = hc.Config(options)
        stride = 1
        if options.stride is not None:
            stride = options.stride
        filter = 3
        if options.filter is not None:
            filter = options.filter
        padding = 1
        if options.padding is not None:
            padding = options.padding

        dilation = 1

        layer = nn.Conv2d(options.input_channels or self.current_size.channels, channels, filter, stride, padding = (padding, padding))
        self.nn_init(layer, options.initializer)
        h, w = self.conv_output_shape((self.current_size.height, self.current_size.width), filter, stride, padding, dilation)
        self.current_size = LayerShape(channels, h, w)
        return layer

    def layer_conv1d(self, net, args, options):
        if len(args) > 0:
            channels = args[0]
        else:
            channels = self.current_size.channels
        print("Options:", options)
        options = hc.Config(options)
        stride = options.stride or 1
        fltr = options.filter or 3
        dilation = 1

        padding = 1
        if options.padding is not None:
            padding = options.padding

        layers = [nn.Conv1d(options.input_channels or self.current_size.channels, channels, fltr, stride, padding = padding)]
        self.nn_init(layers[-1], options.initializer)
        h, _ = self.conv_output_shape((self.current_size.height, self.current_size.height), options.filter or 3, stride, padding, 1)
        self.current_size = LayerShape(channels, h)
        return nn.Sequential(*layers)


    def layer_conv3d(self, net, args, options):
        if len(args) > 0:
            channels = args[0]
        else:
            channels = self.current_size.channels
        options = hc.Config(options)
        stride = options.stride or 1
        fltr = options.filter or 3
        dilation = 1

        padding = options.padding or 1#self.get_same_padding(self.current_width, self.current_width, stride, dilation)
        if options.padding0:
            padding = [options.padding0, padding, padding]
        if options.stride0:
            stride = [options.stride0, stride, stride]
        else:
            stride = [stride, stride, stride]

        layers = [nn.Conv3d(options.input_channels or self.current_size.channels, channels, fltr, stride, padding = padding)]
        self.nn_init(layer, options.initializer)
        self.current_size = LayerShape(frames, channels, self.current_size.height // stride[1], self.current_size.width // stride[2]) #TODO this doesn't work, what is frames? Also chw calculation like conv2d
        return nn.Sequential(*layers)

    def layer_linear(self, net, args, options):
        options = hc.Config(options)
        shape = [int(x) for x in str(args[0]).split("*")]
        bias = True
        if options.bias == False:
            bias = False
        output_size = 1
        for dim in shape:
            output_size *= dim
        layers = []
        if len(self.current_size.dims) != 1:
            layers += [nn.Flatten()]

        layers += [nn.Linear(options.input_size or self.current_size.size(), output_size, bias=bias)]
        self.nn_init(layers[-1], options.initializer)
        self.current_size = LayerShape(*list(reversed(shape)))
        if len(shape) != 1:
            layers.append(Reshape(*self.current_size.dims))

        return nn.Sequential(*layers)

    def layer_modulated_conv2d(self, net, args, options):
        channels = self.current_size.channels
        if len(args) > 0:
            channels = args[0]
        method = "conv"
        if len(args) > 1:
            method = args[1]
        upsample = method == "upsample"
        downsample = method == "downsample"

        demodulate = True
        if options.demodulate == False:
            demodulate = False

        filter = 3
        if options.filter:
            filter = options.filter

        lr_mul = 1.0
        if options.lr_mul:
            lr_mul = options.lr_mul
        input_channels = self.current_size.channels
        if options.input_channels:
            input_channels = options.input_channels

        result = ModulatedConv2d(input_channels, channels, filter, self.layer_output_sizes['w'].size(), upsample=upsample, demodulate=demodulate, downsample=downsample, lr_mul=lr_mul)

        if upsample:
            self.current_size = LayerShape(channels, self.current_size.height * 2, self.current_size.width * 2)
        elif downsample:
            self.current_size = LayerShape(channels, self.current_size.height // 2, self.current_size.width // 2)
        return result

    def layer_module(self, net, args, options):
        klass = GANComponent.lookup_function(None,"function:__main__."+args[0])
        instance = klass(self.gan, net, args, options, self.current_size)
        self.current_size = instance.layer_shape(self.current_size)
        return instance

    def layer_blur(self, net, args, options):
        blur_kernel=[1, 3, 3, 1]
        kernel_size=3
        factor = 2
        p = (len(blur_kernel) - factor) - (kernel_size - 1)
        pad0 = (p + 1) // 2 + factor - 1
        pad1 = p // 2 + 1

        return Blur(blur_kernel, pad=(pad0, pad1), upsample_factor=factor)

    def layer_reshape(self, net, args, options):
        dims_args = [int(x) for x in args[0].split("*")]
        dims = list(reversed(dims_args))
        self.current_size = LayerShape(*dims)
        return Reshape(*dims)

    def layer_adaptive_avg_pool(self, net, args, options):
        self.current_size = LayerShape(self.current_size.channels, self.current_size.height // 2, self.current_size.width // 2)
        return nn.AdaptiveAvgPool2d([self.current_size.height, self.current_size.width])

    def layer_adaptive_avg_pool1d (self, net, args, options):
        self.current_size = LayerShape(self.current_size.channels, self.current_size.height // 2)
        return nn.AdaptiveAvgPool1d(self.current_size.height)

    def layer_avg_pool(self, net, args, options):
        self.current_size = LayerShape(self.current_size.channels, self.current_size.height // 2, self.current_size.width // 2)
        return nn.AvgPool2d(2, 2)

    def layer_adaptive_avg_pool3d(self, net, args, options):
        frames = 4 #TODO
        self.current_size = LayerShape(frames, self.current_size.channels, self.current_size.height // 2, self.current_size.width // 2)
        return nn.AdaptiveAvgPool3d([self.current_size.frames, self.current_size.height, self.current_size.width]) #TODO looks wrong

    def layer_instance_norm(self, net, args, options):
        options = hc.Config(options)
        affine = True
        if options.affine == False:
            affine = False
        return nn.InstanceNorm2d(self.current_size.channels, affine=affine)

    def layer_instance_norm1d(self, net, args, options):
        options = hc.Config(options)
        affine = True
        if options.affine == False:
            affine = False
        return nn.InstanceNorm1d(self.current_size.channels, affine=affine)


    def layer_instance_norm3d(self, net, args, options):
        options = hc.Config(options)
        affine = True
        if options.affine == False:
            affine = False
        return nn.InstanceNorm3d(self.current_size.channels, affine=affine)

    def layer_batch_norm(self, net, args, options):
        return nn.BatchNorm2d(self.current_size.channels)

    def layer_batch_norm1d(self, net, args, options):
        return nn.BatchNorm1d(self.current_size.size())

    def get_conv_options(self, config, options):
        stride = options.stride or self.ops.config_option("stride", [1,1])
        fltr = options.filter or self.ops.config_option("filter", [3,3])
        avg_pool = options.avg_pool or self.ops.config_option("avg_pool", [1,1])

        if type(stride) != type([]):
            stride = [stride, stride]

        if type(avg_pool) != type([]):
            avg_pool = [avg_pool, avg_pool]

        if type(fltr) != type([]):
            fltr = [fltr, fltr]
        return stride, fltr, avg_pool


    def layer_deconv(self, net, args, options):
        if len(args) > 0:
            channels = args[0]
        else:
            channels = self.current_size.channels
        options = hc.Config(options)
        filter = 4 #TODO
        if options.filter:
            filter = options.filter
        stride = 2
        if options.stride:
            stride = options.stride
        padding = 1
        if options.padding:
            padding = options.padding
        layer = nn.ConvTranspose2d(options.input_channels or self.current_size.channels, channels, filter, stride, padding)
        self.nn_init(layer, options.initializer)
        self.current_size = LayerShape(channels, self.current_size.height * 2, self.current_size.width * 2)
        return layer

    def layer_pad(self, net, args, options):
        options = hc.Config(options)

        return nn.ZeroPad2d((args[0], args[1], args[2], args[3]))

    def layer_pixel_norm(self, net, args, options):
        return PixelNorm()

    def layer_pretrained(self, net, args, options):
        model = getattr(torchvision.models, args[0])(pretrained=True)
        model.train(True)
        if options.layer:
            layers = list(model.children())[:options.layer]
            if options.sublayer:
                layers[-1] = nn.Sequential(*layers[-1][:options.sublayer])
        else:
            layers = [model]
            print("List of pretrained layers:", layers)
            raise ValidationException("layer=-1 required for pretrained, sublayer=-1 optional.  Layers outputted above.")
        return nn.Sequential(*layers)

    def layer_resize_conv(self, net, args, options):
        return self.layer_resize_conv2d(net, args, options)

    def layer_resize_conv2d(self, net, args, options):
        options = hc.Config(options)
        channels = args[0]

        w = options.w or self.current_size.width * 2
        h = options.h or self.current_size.height * 2
        layers = [nn.Upsample((h, w), mode="bilinear"),
                nn.Conv2d(options.input_channels or self.current_size.channels, channels, options.filter or 3, 1, 1)]
        self.nn_init(layers[-1], options.initializer)
        self.current_size = LayerShape(channels, h, w)
        return nn.Sequential(*layers)

    def layer_resize_conv1d(self, net, args, options):
        options = hc.Config(options)
        channels = args[0]
        h = options.h or self.current_size.height * 2

        padding = 1
        if options.padding is not None:
            padding = options.padding

        layers = [nn.Upsample((h)),
                nn.Conv1d(options.input_channels or self.current_size.channels, channels, options.filter or 3, 1, padding=padding)]
        self.nn_init(layers[-1], options.initializer)
        h, _ = self.conv_output_shape((self.current_size.height, self.current_size.height), options.filter or 3, 1, padding, 1)
        self.current_size = LayerShape(channels, h)
        return nn.Sequential(*layers)

    def layer_scaled_conv2d(self, net, args, options):
        channels = self.current_size.channels
        if len(args) > 0:
            channels = args[0]
        method = "conv"
        if len(args) > 1:
            method = args[1]
        upsample = method == "upsample"
        downsample = method == "downsample"

        demodulate = True
        if options.demodulate == False:
            demodulate = False

        filter = 3
        if options.filter:
            filter = options.filter

        lr_mul = 1.0
        if options.lr_mul:
            lr_mul = options.lr_mul
        input_channels = self.current_size.channels
        if options.input_channels:
            input_channels = options.input_channels

        result = ScaledConv2d(input_channels, channels, filter, 0, upsample=upsample, demodulate=demodulate, downsample=downsample, lr_mul=lr_mul)
        self.nn_init(result, options.initializer)

        if upsample:
            self.current_size = LayerShape(channels, self.current_size.height * 2, self.current_size.width * 2)
        else:
            self.current_size = LayerShape(channels, self.current_size.height - 2, self.current_size.width - 2)
        return result

    def layer_split(self, net, args, options):
        options = hc.Config(options)
        split_size = args[0]
        select = args[1]
        dim = -1
        if options.dim:
            dim = options.dim
        #TODO better validation
        #TODO increase dim options
        if dim == -1:
            dims = list(self.current_size.dims).copy()
            dims[0] = split_size
            if (select + 1) * split_size > self.current_size.channels:
                dims[0] = self.current_size.channels % split_size
            self.current_size = LayerShape(*dims)
        return NoOp()

    def layer_subpixel(self, net, args, options):
        options = hc.Config(options)
        channels = args[0]

        layers = [nn.Conv2d(options.input_channels or self.current_size.channels, channels*4, options.filter or 3, 1, 1), nn.PixelShuffle(2)]
        self.nn_init(layers[0], options.initializer)
        self.current_size = LayerShape(channels, self.current_size.height * 2, self.current_size.width * 2)
        return nn.Sequential(*layers)

    def layer_latent(self, net, args, options):
        self.current_size = LayerShape(self.gan.latent.current_input_size)
        return NoOp()

    def layer_layer(self, net, args, options):
        if args[0] in self.layer_output_sizes:
            self.current_size = self.layer_output_sizes[args[0]]
        return NoOp()

    def layer_linformer(self, net, args, options):
        model = Linformer(
                input_size = self.current_size.size(),
                channels = self.current_size.height # TODO wtf
        )
        return model

    def layer_vae(self, net, args, options):
        self.vae = Variational(self.current_size.channels)
        return self.vae

    def layer_multi_head_attention(self, net, args, options):
        output_size = self.current_size.size()
        if len(args) > 0:
            output_size = args[0]
        layer = MultiHeadAttention(self.current_size.size(), output_size, heads=options.heads or 4)
        self.current_size = LayerShape(output_size)
        self.nn_init(layer.o, options.initializer)
        self.nn_init(layer.h, options.initializer)
        self.nn_init(layer.g, options.initializer)
        self.nn_init(layer.f, options.initializer)
        return layer

    def layer_attention(self, net, args, options):
        layer = Attention(self.current_size.channels)
        self.nn_init(layer.v, options.initializer)
        self.nn_init(layer.h, options.initializer)
        self.nn_init(layer.g, options.initializer)
        self.nn_init(layer.f, options.initializer)
        return layer

    def layer_norm(self, net, args, options):
        affine = True
        if options.affine == False:
            affine = False

        return nn.LayerNorm(self.current_size.dims, elementwise_affine=affine)

    def layer_learned_noise(self, net, args, options):
        return LearnedNoise(*([self.gan.batch_size(), *self.current_size.dims]))

    def layer_adaptive_instance_norm(self, net, args, options):
        return AdaptiveInstanceNorm(self.layer_output_sizes['w'].size(), self.current_size.channels, equal_linear=options.equal_linear)

    def layer_flatten(self, net, args, options):
        self.current_size = LayerShape(self.current_size.size())
        return nn.Flatten()

    def layer_zeros_like(self, net, args, options):
        return Zeros(self.gan.latent.sample().shape)

    def nn_init(self, layer, initializer_option):
        if initializer_option is None:
            return
        if type(initializer_option) == pyparsing.ParseResults and type(initializer_option[0]) == hypergan.parser.Pattern:
            args = [initializer_option[0].layer_name] + initializer_option[0].args
            options = hc.Config(initializer_option[0].options)
        else:
            args = [initializer_option]
            options = hc.Config({})

        layer_data = layer.weight.data

        if args[0] == "uniform":
            a = float(args[1])
            b = float(args[2])
            nn.init.uniform_(layer_data, a, b)
        elif args[0] == "normal":
            mean = float(args[1])
            std = float(args[2])
            nn.init.normal_(layer_data, mean, std)
        elif args[0] == "constant":
            val = float(args[1])
            nn.init.constant_(layer_data, val)
        elif args[0] == "ones":
            nn.init.ones_(layer_data)
        elif args[0] == "zeros":
            nn.init.zeros_(layer_data)
        elif args[0] == "eye":
            nn.init.eye_(layer_data)
        elif args[0] == "dirac":
            nn.init.dirac_(layer_data)
        elif args[0] == "xavier_uniform":
            gain = nn.init.calculate_gain(options.gain or "relu")
            nn.init.xavier_uniform_(layer_data, gain=gain)
        elif args[0] == "xavier_normal":
            gain = nn.init.calculate_gain(options.gain or "relu")
            nn.init.xavier_normal_(layer_data, gain=gain)
        elif args[0] == "kaiming_uniform":
            a = 0 #TODO wrong
            nn.init.kaiming_uniform_(layer_data, mode=(options.mode or "fan_in"), nonlinearity=options.gain or "relu")
        elif args[0] == "kaiming_normal":
            a = 0 #TODO wrong
            nn.init.kaiming_normal_(layer_data, mode=(options.mode or "fan_in"), nonlinearity=options.gain or "relu")
        elif args[0] == "orthogonal":
            if "gain" in options:
                gain = nn.init.calculate_gain(options["gain"])
            else:
                gain = 1
            nn.init.orthogonal_(layer_data, gain=gain)
        else:
            print("Warning: No initializer found for " + args[0])
        if "gain" in options:
            layer_data.mul_(nn.init.calculate_gain(options["gain"]))
        return NoOp()


    def forward(self, input, context={}):
        for module, parsed, layer_shape in zip(self.net, self.parsed_layers, self.layer_shapes):
            try:
                options = parsed.parsed_options
                args = parsed.args
                layer_name = parsed.layer_name
                name = options.name
                if isinstance(module, hg.Layer):
                    input = module(input, context)
                elif layer_name == "adaptive_instance_norm":
                    input = module(input, context['w'])
                elif layer_name == "ez_norm":
                    input = module(input, context['w'])
                elif layer_name == "split":
                    input = torch.split(input, args[0], options.dim or -1)[args[1]]
                elif layer_name == "layer":
                    input = context[args[0]]
                elif layer_name == "latent":
                    input = self.gan.latent.z#sample()
                elif layer_name == "modulated_conv2d":
                    input = module(input, context['w'])
                elif layer_name == "pretrained":
                    in_zero_one = (input + self.const_one) / self.const_two
                    mean = torch.as_tensor([0.485, 0.456, 0.406], device='cuda:0').view(1, 3, 1, 1)
                    std = torch.as_tensor([0.229, 0.224, 0.225], device='cuda:0').view(1, 3, 1, 1)

                    input = module(input.clone().sub_(mean).div_(std))
                else:
                    input = module(input)
                if self.gan.steps == 0:
                    size = LayerShape(*list(input.shape[1:]))
                    if size.squeeze_dims() != layer_shape.squeeze_dims():
                        print("Error: Size error on", layer_name)
                        print("Error: Expected output size", layer_shape.dims)
                        print("Error: Actual output size", size.dims)
                        raise "Layer size error, cannot continue"
                    else:
                        print("Sizes as expected", input.shape[1:], layer_shape.dims)
                if name is not None:
                    context[name] = input
            except:
                raise ValidationException("Error on " + parsed.layer_defn + " - input size " + ",".join([str(x) for x in input.shape]))
        self.sample = input
        return input

    def set_trainable(self, flag):
        for p in (set(list(self.parameters())) - self.untrainable_parameters):
            p.requires_grad = flag



"""
Opens a window that displays an image.
Usage:

    from viewer import GlobalViewer
    GlobalViewer.update(image)

Delays loading Gtk and friends until enable() is called.
"""
from hypergan.threaded_tk_viewer import ThreadedTkViewer

GlobalViewer = ThreadedTkViewer()



import numpy as np
import hyperchamber as hc
import torch.nn as nn

from .base_generator import BaseGenerator

class DCGANGenerator(BaseGenerator):

    def required(self):
        return []

    def create(self):
        self.linear = nn.Sequential(
                nn.Linear(100, 4*4*512),
                nn.ReLU()
        )
        self.net = nn.Sequential(
            nn.ConvTranspose2d(512, 256, 4, 2, 1),
            nn.BatchNorm2d(256),
            nn.ReLU(),
            nn.ConvTranspose2d(256, 128, 4, 2, 1),
            nn.BatchNorm2d(128),
            nn.ReLU(),
            nn.ConvTranspose2d(128, 64, 4, 2, 1),
            nn.BatchNorm2d(64),
            nn.ReLU(),
            nn.ConvTranspose2d(64, 3, 4, 2, 1),
            nn.Tanh()
        )

    def forward(self, x):
        lin = self.linear(x).view(self.gan.batch_size(), 512, 4, 4)
        net = self.net(lin)
        return net.view(self.gan.batch_size(),3,64,64)



from hypergan.gan_component import GANComponent

class BaseGenerator(GANComponent):

    def __init__(self, gan, config, input=None):
        GANComponent.__init__(self, gan, config)
        self.input = input

    """
        Superclass for all Generators.  Provides some common functionality.
    """
    def add_progressive_enhancement(self, net):
        ops = self.ops
        gan = self.gan
        config = self.config
        if config.progressive_enhancement:
            split = ops.slice(net, [0, 0, 0, 0], [-1, -1, -1, gan.channels()])
            if config.final_activation:
                split = config.final_activation(split)
            print("[generator] adding progressive enhancement", split)
            gan.skip_connections.set('progressive_enhancement', split)


    def project_from_prior(self, primes, net, initial_depth, type='linear', name='prior_projection'):
        ops = self.ops
        net = ops.reshape(net, [ops.shape(net)[0], -1])
        new_shape = [ops.shape(net)[0], primes[0], primes[1], initial_depth]
        net = ops.linear(net, initial_depth*primes[0]*primes[1])
        print("projection ", net)
        net = ops.reshape(net, new_shape)
        return net





from ..gan_component import ValidationException
from .base_distribution import BaseDistribution
from hypergan.gan_component import ValidationException, GANComponent
from torch.autograd import Variable
from torch.autograd import grad as torch_grad
from torch.distributions import uniform
import hyperchamber as hc
import numpy as np
import torch

class OptimizeDistribution(BaseDistribution):
    def __init__(self, gan, config):
        BaseDistribution.__init__(self, gan, config)
        klass = GANComponent.lookup_function(None, self.config['source']['class'])
        self.source = klass(gan, config['source'])
        self.current_channels = config['source']["z"]
        self.current_width = 1
        self.current_height = 1
        self.current_input_size = config['source']["z"]
        self.z = self.source.z
        self.hardtanh = torch.nn.Hardtanh()
        self.relu = torch.nn.ReLU()
        self.z_var = None

    def create(self):
        pass

    def next(self):
        sample = self.source.next()
        if self.z_var is None:
            self.z_var = Variable(sample, requires_grad=True).cuda()
        defn = self.config.optimizer.copy()
        klass = GANComponent.lookup_function(None, defn['class'])
        del defn["class"]
        self.optimizer = klass([self.z_var], **defn)
        z = self.z_var
        z.data.copy_(sample)

        z.grad = torch.zeros_like(z)
        for i in range(self.config.steps or 1):
            self.optimizer.zero_grad()
            fake = self.gan.discriminator(self.gan.generator(self.hardtanh(z))).mean()
            real = self.gan.discriminator(self.gan.inputs.sample).mean()
            loss = self.gan.loss.forward_adversarial_norm(real, fake)
            if loss == 0.0:
                if self.config.verbose:
                    print("[optimize distribution] No loss")
                break
            z_move = torch_grad(outputs=loss, inputs=z, retain_graph=True, create_graph=True)
            z_change = z_move[0].abs().mean()
            if self.config.z_change_threshold or self.config.verbose:
                if i == 0:
                    first_z_change = z_change
            z._grad.copy_(z_move[0])
            self.optimizer.step()
            if self.config.verbose:
                print("[optimize distribution]", i, "loss", loss.item(), "mean movement", (z-sample).abs().mean().item(), (z_change/first_z_change).item())
            if self.config.z_change_threshold and z_change/first_z_change < self.config.z_change_threshold:
                if self.config.info:
                    print("[optimize distribution] z_change_threshold steps", i, "loss", loss.item(), "mean movement", (z-sample).abs().mean().item(), (z_change/first_z_change).item())
                break
            if self.config.loss_threshold and loss < self.config.loss_threshold:
                if self.config.info:
                    print("[optimize distribution] loss_threshold steps", i, "loss", loss.item(), "mean movement", (z-sample).abs().mean().item(), (z_change/first_z_change).item())
                break

        if self.config.info and i == self.config.steps-1:
            print("[optimize distribution] steps_threshold steps", i, "loss", loss.item(), "mean movement", (z-sample).abs().mean().item())
        self.instance = z
        return z



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from tensorflow.python.ops import control_flow_ops
from tensorflow.python.ops import math_ops
from tensorflow.python.ops import state_ops
from tensorflow.python.framework import ops
from tensorflow.python.training import optimizer
import hyperchamber as hc
import inspect

class SOSOptimizer(optimizer.Optimizer):
  def __init__(self, learning_rate=0.001, p=0.1, gan=None, config=None, use_locking=False, name="SOSOptimizer", optimizer=None, alpha=1):
    super().__init__(use_locking, name)
    self._alpha = alpha
    self.gan = gan
    self.config = config
    self._lr_t = learning_rate
    self.optimizer = self.gan.create_optimizer(optimizer)
 
  def _prepare(self):
    super()._prepare()
    self.optimizer._prepare()

  def _create_slots(self, var_list):
    super()._create_slots(var_list)
    self.optimizer._create_slots(var_list)

  def _apply_dense(self, grad, var):
    return self.optimizer._apply_dense(grad, var)

  def apply_gradients(self, grads_and_vars, global_step=None, name=None):
    all_vars = [ v for _,v in grads_and_vars]
    d_vars = []
    g_vars = []
    all_grads = [ g for g, _ in grads_and_vars ]
    for grad,var in grads_and_vars:
        if var in self.gan.d_vars():
            d_vars += [var]
        elif var in self.gan.g_vars():
            g_vars += [var]
        else:
            raise("Couldn't find var in g_vars or d_vars")

    with ops.init_scope():
        [self._get_or_make_slot(v, v, "restore", self._name) for v in all_vars]
        self.optimizer._create_slots([v for g,v in grads_and_vars])
        for name in self.optimizer.get_slot_names():
            for var in self.optimizer.variables():
                self._zeros_slot(var, "restore", self._name)

    self._prepare()
    restore_slots = [self.get_slot(v, "restore") for v in all_vars]
    restore_vars = all_vars
    for name in self.optimizer.get_slot_names():
        for var in self.optimizer.variables():
            restore_vars += [var]
            restore_slots += [self._zeros_slot(var, "restore", self._name)]


    d_grads = all_grads[:len(d_vars)]
    g_grads = all_grads[len(d_vars):]
    dc_grads = [tf.reduce_sum(tf.square(d)) for d in d_grads]
    gc_grads = [tf.reduce_sum(tf.square(g)) for g in g_grads]
    l1 = self.gan.trainer.d_loss# + lookahead_l2
    l2 = self.gan.trainer.g_loss# + lookahead_l1
    #h11 = tf.gradients(d_grads, d_vars) + tf.gradients(d_grads, g_vars, stop_gradients=g_vars)
    h11 = tf.gradients(d_grads, d_vars) + [tf.zeros_like(g) for g in g_vars]
    h12 = tf.gradients(gc_grads, d_vars) + [tf.zeros_like(g) for g in g_vars]
    h21 = [tf.zeros_like(d) for d in d_vars] + tf.gradients(dc_grads, g_vars)
    #h22 = tf.gradients(g_grads, d_vars) + tf.gradients(g_grads, g_vars, stop_gradients=d_vars)
    h22 = [tf.zeros_like(d) for d in d_vars] + tf.gradients(g_grads, g_vars)
    #h22 = [tf.zeros_like(d) for d in all_vars]
    h11 = [ tf.zeros_like(_dg) if ddg is None else _dg for ddg, _dg in zip(all_vars, h11) ]
    h12 = [ tf.zeros_like(_dg) if ddg is None else _dg for ddg, _dg in zip(all_vars, h12) ]
    h21 = [ tf.zeros_like(_dg) if ddg is None else _dg for ddg, _dg in zip(all_vars, h21) ]
    h22 = [ tf.zeros_like(_dg) if ddg is None else _dg for ddg, _dg in zip(all_vars, h22) ]
    h21 = [ -_dg for _dg in h12 ]
    h22 = [ -_dg for _dg in h12 ]
    __h11 = [ tf.reduce_sum(_h11) for _h11 in h11 ]
    __h12 = [ tf.reduce_sum(_h12) for _h12 in h12 ]
    __h21 = [ tf.reduce_sum(_h21) for _h21 in h21 ]
    __h22 = [ tf.reduce_sum(_h22) for _h22 in h22 ]
    #h12_metric = self.gan.ops.squash(sum(h12))
    h11_metric = self.gan.ops.squash(sum(__h11))
    self.gan.add_metric('h11', h11_metric)
    h12_metric = self.gan.ops.squash(sum(__h12))
    self.gan.add_metric('h12', h12_metric)
    h21_metric = self.gan.ops.squash(sum(__h21))
    self.gan.add_metric('h21', h21_metric)
    h22_metric = self.gan.ops.squash(sum(__h22))
    self.gan.add_metric('h22', h22_metric)
    ho = []
    for _h12, _h21 in zip(h12, h21):
        zero = tf.zeros_like(_h12)
        shape = self.gan.ops.shape(_h12)
        #_ho = tf.stack([zero, _h11, _h22, zero])
        _ho = tf.stack([tf.stack([zero, _h12]),tf.stack([_h21, zero])])
        ho.append(_ho)

   
    eps = []
    m = tf.constant(0.0)
    for _h12, _h21, _grads, _ho in zip(h12, h21, all_grads, ho):
        # (I - alpha*Ho)
        Eo = 2.0 * _grads - \
             self._alpha*_h21 * _grads -\
             self._alpha*_h12 * _grads
        m += tf.reduce_sum(Eo)
        eps += [ Eo ]

    self.gan.add_metric('m', m)
    new_grads = eps
    new_grads_and_vars = list(zip(new_grads, all_vars)).copy()

    diag = []
    ho_t = []


    op1 = tf.group(*[tf.assign(w, v) for w,v in zip(restore_slots, restore_vars)]) # store variables
    with tf.get_default_graph().control_dependencies([op1]):
        #lookahead = DepthOptimizer(learning_rate=0.001, decay=1.0, gan=None, config=None, use_locking=False, name="depthOptimizer", optimizer=self.optimizer, depth=1)
        #op2 = lookahead.apply_gradients(eps)
        op2 = self.optimizer.apply_gradients(new_grads_and_vars, global_step=global_step, name=name)
        with tf.get_default_graph().control_dependencies([op2]):
            new_grads2 = tf.gradients(self.gan.trainer.d_loss, d_vars) + tf.gradients(self.gan.trainer.g_loss, g_vars)
            new_grads3 = []
            p1s = tf.constant(0.0)
            p2s = tf.constant(0.0)
            p1_masks = tf.constant(0.0)
            p2_masks = tf.constant(0.0)
            p1v2s = tf.constant(0.0)
            ps = tf.constant(0.0)
            for _grad1, _grad2, _ho in zip(eps, new_grads2, ho):
                axis = [i for i in range(len(self.gan.ops.shape(_ho)))]
                _X = (_ho[1][0] + _ho[0][1])*_grad2
                _Xr = tf.reshape(_X, [-1])
                _grad1r = tf.reshape(_grad1, [-1])
                angle = tf.tensordot(-self._alpha * _Xr, _grad1r, 1)
                p1_mask = (tf.sign(angle) + 1.0) / 2.0
                aeo = - self.config.a * (tf.square(tf.norm(_grad1)))/(angle+1e-8)
                p1 = tf.minimum(1.0, aeo)
                p1 = p1_mask + (1.0-p1_mask) * p1
                p2_mask = (tf.sign((tf.norm(_grad1)) - self.config.b) + 1.0) / 2.0
                p2 = tf.square(tf.norm(_grad1)) * p2_mask + (1.0-p2_mask)
                p = tf.minimum(p1, p2)
                p1v2 = tf.reduce_sum(tf.sign(p1-p2))
                ps+=tf.reduce_sum(p)
                p1s+=tf.reduce_sum(p1)
                p2s+=tf.reduce_sum(p2)
                p1v2s+=tf.reduce_sum(p1v2)
                p1_masks+=tf.reduce_sum(p1_mask)
                p2_masks+=tf.reduce_sum(p2_mask)

                if self.config.lola:
                    new_grads3 += [_grad1 - _X * self._alpha]
                else:
                    new_grads3 += [_grad1 - p * _X * self._alpha]

            self.gan.add_metric("p1m",p1_masks)
            self.gan.add_metric("p2m",p2_masks)
            self.gan.add_metric("p1v2",p1v2s)
            self.gan.add_metric("p",ps)
            self.gan.add_metric("p1",p1s)
            self.gan.add_metric("p2",p2s)
            new_grads_and_vars3 = list(zip(new_grads3, all_vars)).copy()
            op3 = tf.group(*[tf.assign(w, v) for w,v in zip(restore_vars, restore_slots)]) # store variables
            with tf.get_default_graph().control_dependencies([op3]):
                return self.optimizer.apply_gradients(new_grads_and_vars3, global_step=global_step, name=name)

  
  def _apply_sparse(self, grad, var):
    raise NotImplementedError("Sparse gradient updates are not supported.")



import hyperchamber as hc
import torch.nn as nn
import inspect
import os

from .base_discriminator import BaseDiscriminator

class DCGANDiscriminator(BaseDiscriminator):

    def required(self):
        return []

    def create(self):
        self.net = nn.Sequential(
                nn.Conv2d(3, 64, 3, 2, 1),
                nn.ReLU(),
                nn.Conv2d(64, 128, 3, 2, 1),
                nn.ReLU(),
                nn.Conv2d(128, 256, 3, 2, 1),
                nn.ReLU(),
                nn.Conv2d(256, 512, 3, 2, 1),
                nn.ReLU()
        )
        self.linear = nn.Linear(4*4*512, 1)

    def forward(self, x):
        net = self.net(x).view(self.gan.batch_size(), -1)
        return self.linear(net).view(self.gan.batch_size(),1)
